
import numpy as np
import pandas as pd
from ..._arithmetic._funcs._SequenceManipulation import _SequenceManipulation

def _add_protospacer_sequence(df, chromosome_seq, strand):

    """Add the sequence of the protospacer and the PAM to the df."""

    df["PAM_loci"] = df.PAM_loci.astype(int)

    pam_seqs = np.array([])
    guide_seqs = np.array([])

    if strand == "+":

        for i in df['PAM_loci']:
            pam_seqs = np.append(pam_seqs, chromosome_seq[i : i + 3])
        for i in df.index:
            guide_seqs = np.append(
                guide_seqs, chromosome_seq[df.guide_begin[i] : df.guide_end[i]]
            )

    elif strand == "-":

        for i in df['PAM_loci']:            
            pam_seq = _SequenceManipulation(chromosome_seq[i - 3 : i])
            pam_seqs = np.append(pam_seqs, pam_seq.reverse_complement())
                
        for i in df.index:
            guide_seq = _SequenceManipulation(chromosome_seq[df.guide_end[i] : df.guide_begin[i]])
            guide_seqs = np.append(guide_seqs, guide_seq.reverse_complement())
    else:
        print("Strand unassigned. Interrupting.")
        return

    df["PAM"] = pam_seqs
    df["protospacer"] = guide_seqs

    return df

def _annotate_protospacer(df, chromosome_seq):

    """
    Annotate protospacers and PAM sequence. 
    
    Annotate df with information about guide protospacer loci and sequence.
    Splits df into forward and reverse strand, adds loci of protospacer then recombines as before.
    """
    forw_df = df.loc[df.strand == "+"]
    revr_df = df.loc[df.strand == "-"]

    forw_df["guide_begin"] = (forw_df.PAM_loci - 20).astype(int)
    forw_df["guide_end"] = (forw_df.PAM_loci).astype(int)
    forw_df = _add_protospacer_sequence(forw_df, chromosome_seq, strand="+")

    revr_df["guide_begin"] = (revr_df.PAM_loci + 20).astype(int)
    revr_df["guide_end"] = (revr_df.PAM_loci).astype(int)
    revr_df = _add_protospacer_sequence(revr_df, chromosome_seq, strand="-")

    df = pd.concat([forw_df, revr_df]).sort_values("PAM_loci").reset_index(drop=True)

    return df


class _SequenceManipulation:
    
    """
    Get the complement or reverse compliment of a DNA sequence.
    
    Parameters:
    -----------
    sequence
    
    Returns:
    --------
    complement_sequence
    
    reverse_sequence
    
    reverse_complement_sequence
    
    Notes:
    ------
    (1) No dependencies required. Pure python. 
    """
    
    def __init__(self, sequence):
        
        self.sequence = sequence
        self.ComplimentDict = {'C':'G', 'G':'C', 'T':'A', 'A':'T', 'N':'N'}
        self.complement_sequence = ""
        self.reverse_complement_sequence = ""
        
    def complement(self):
        
        """
        Get the complement of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        complement_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python. 
        """
        for nucleotide in self.sequence:
            self.complement_sequence += self.ComplimentDict[nucleotide]
        
    
        return self.complement_sequence
        
    def reverse(self):
        
        """
        Get the reverse of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        reverse_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python. 
        """
        
        self.reverse_sequence = self.sequence[::-1]
        
    def reverse_complement(self):
        
        """
        Get the reverse complement of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        reverse_complement_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python. 
        """
        
        if not self.complement_sequence:
            for nucleotide in self.sequence:
                self.complement_sequence += self.ComplimentDict[nucleotide]
        
        self.reverse_complement_sequence = self.complement_sequence[::-1]
        return self.reverse_complement_sequence



#import vintools as v
import numpy as np
import matplotlib.pyplot as plt

def _cds_plot_construction():
    NotImplemented
    # plot = v.pl.ScatterPlot()
    # plot.construct_layout(figsize_width=4, figsize_height=0.5, nplots=1)
    # plot.style(grid=False)
    # ax = plot.AxesDict[0][0]

    # spines = v.pl.ax_spines(ax)
    # spines.delete(["left"])

    # return plot, ax


def _plot_cds(df, title, exon_spacing=100):

    """"""
    NotImplemented
    # c = v.pl.share_seq()["colors"]
    # c = np.repeat(c, 3).values

    # plot, ax = _cds_plot_construction()

    # floating_start = 0
    # exon_centers = []
    # for i in range(len(df)):
    #     stop = floating_start + df["exon_length"].iloc[i]
    #     exon_centers.append(np.mean([stop, floating_start]))
    #     ax.hlines(0.1, floating_start, stop, color=c[i], linewidth=10)
    #     floating_start = stop + exon_spacing

    # ax.set_ylabel(title, rotation=0, y=0.05)

    # ax.set_yticks([])
    # ax.set_ylim(0, 1)
    # xticks = ax.set_xticks(exon_centers)
    # xticks = ax.set_xticklabels(range(len(df)), fontsize=8)
    # plt.savefig(title + ".pdf")



from ..._utilities._funcs._SequenceManipulation import _SequenceManipulation
from Bio import SeqIO
import pandas as pd
import numpy as np
import regex


def _get_seq_one_chr(ref_seq_path, query_chr):

    """
    Get a specific chromosome sequence from a reference genome. Also report the length of that sequence.
    Parameters:
    -----------
    ref_seq_path
        Path to a reference genome fasta file.
    query_chr
        Chromosome sequence to isolate
        type: str
    Returns:
    --------
    chromosome_reference_seq
    
    len(chromosome_reference_seq)
    """

    for record in SeqIO.parse(ref_seq_path, "fasta"):

        if record.description.split()[0] == query_chr:
            chromosome_reference_seq = str(record.seq)

            return chromosome_reference_seq, len(chromosome_reference_seq)

def _adjust_reverse_strand_coords_to_forward(
    reverse_seq_pam_match_coords, chromosome_length
):

    """
    Given a pam and a sequece, find sgRNAs within that sequence.
    Parameters:
    -----------
    reverse_seq_pam_match_coords
        Numpy array contianing loci of the first base in the PAM sequence. Numbers start right to left (5' to 3')
    Returns:
    --------
    reverse_seq_pam_match_coords_adjusted
        Numpy array contianing loci of the first base in the PAM sequence. Numbers start left to right (3' to 5')
    """

    return (chromosome_length - reverse_seq_pam_match_coords)


def _find_pam_loci_whole_chromosome(ref_seq_path, query_chr, PAM="NGG"):

    """
    Given a pam and a sequece, find sgRNAs within that sequence.
    Parameters:
    -----------
    seq: DNA sequence
    PAM: PAM sequence. Default=='NGG'
    Returns:
    --------
    forward_PAM_loci
        Numpy array contianing loci of the first base in the PAM sequence. Numbers start left to right (5' to 3')
    reverse_PAM_loci
        Numpy array contianing loci of the first base in the PAM sequence. Numbers start right to left (5' to 3')
    """

    reference_chromosome_sequence, chromosome_length = _get_seq_one_chr(
        ref_seq_path, query_chr=query_chr
    )
    print(
        "Searching",
        str(chromosome_length),
        "bp along",
        query_chr,
        "for PAM sequences...",
    )

    pam = PAM.replace("N", ".")

    forward_seq_pam_match_coords = [
        m.span()[0]
        for m in regex.finditer(
            r"%s" % (pam), reference_chromosome_sequence, overlapped=True
        )
    ]
    
    ref_seq = _SequenceManipulation(reference_chromosome_sequence)
    reverse_complement_seq = ref_seq.reverse_complement()
    
    reverse_complement_seq_pam_match_coords = [
        m.span()[0]
        for m in regex.finditer(r"%s" % (pam), reverse_complement_seq, overlapped=True)
    ]

    forward_seq_pam_match_coords = np.array(forward_seq_pam_match_coords)
    reverse_complement_seq_pam_match_coords = np.array(
        reverse_complement_seq_pam_match_coords
    )

    print(
        str(len(forward_seq_pam_match_coords)),
        "\nPAM matches identified along the forward strand of query chromsome.\n",
    )
    print(
        str(len(reverse_complement_seq_pam_match_coords)),
        "\nPAM matches identified along the reverse strand of query chromsome.\n",
    )
    
    reverse_complement_seq_pam_match_coords_adjusted = (
        _adjust_reverse_strand_coords_to_forward(
            reverse_complement_seq_pam_match_coords, chromosome_length
        )
    )

    return (
        forward_seq_pam_match_coords,
        reverse_complement_seq_pam_match_coords_adjusted,
        reference_chromosome_sequence
    )

def _get_PAMS_in_regions(regions_df, ref_seq_path, chromosome_key="Chromosome", PAM="NGG"):
    
    """
    Looks for PAM loci one chromosome at a time. Any chromosome specified in the peaks bed file is included.
    
    Parameters:
    -----------
    atac_peaks_path
        Path to a bed file of genomic loci. 
    
    ref_seq_path
        Path to a reference genome fasta file.
    PAM
        PAM motif to search (default: NGG)
        type: str
    Returns:
    --------
    """

    forward_pams = []
    reverse_pams = []
    chrom_seqs = []

    for chrom in regions_df[chromosome_key].unique():

        forw_chr, revr_chr, chrom_seq = _find_pam_loci_whole_chromosome(
            ref_seq_path, query_chr=chrom, PAM=PAM
        )
        forward_pams.append(forw_chr)
        reverse_pams.append(revr_chr)
        chrom_seqs.append(chrom_seq)

    return forward_pams, reverse_pams, chrom_seqs

def _query_region_for_PAM(regions_df, ref_seq_path, chromosome_key="Chromosome", PAM="NGG"):

    """
    Looks for PAM loci one chromosome at a time. Any chromosome specified in the peaks bed file is included.

    Parameters:
    -----------
    regions_df
        Path to a bed file of genomic loci.

    ref_seq_path
        Path to a reference genome fasta file.
    PAM
        PAM motif to search (default: NGG)
        type: str
    Returns:
    --------


    """

    forward_pams = []
    reverse_pams = []
    chrom_seqs = []

    for chrom in regions_df[chromosome_key].unique():

        forw_chr, revr_chr, chrom_seq = _find_pam_loci_whole_chromosome(
            ref_seq_path, query_chr=chrom, PAM=PAM
        )
        forward_pams.append(forw_chr)
        reverse_pams.append(revr_chr)
        chrom_seqs.append(chrom_seq)
        
    return forward_pams, reverse_pams, chrom_seqs



# _experimental_design/__init__.py
__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


from ._general._identify_overlapping_fragments import _OverlappingFragments as OverlappingFragments
from ._general._get_chromosome_sequence import _get_chromosome_sequence
from ._general._GTF_Module import _GTF as GTF

from ._general._TargetModule import _ScreeningTarget as Target

from ._general._scan_for_BsmbI import _scan_for_BsmbI



from typing import Dict
import pandas as pd
from ..._framework._ScreenModule import _Screen


def check_indices(df, index=None, colnames=None):
    if index is not None and not (df.index == index).all():
        raise ValueError("df.index does not match index, check your input.")
    if colnames is not None and not (df.columns == colnames).all():
        raise ValueError("df.columns does not match colnames, check your input.")


def read_csvs(
    X_filename: str,
    guides_filename: str = None,
    samples_filename: str = None,
    layers_filenames_dict: Dict[str, str] = None,
    matched_indices: bool = True,
) -> _Screen:
    """Returns Screen object with X, guides, and samples.
    Assumes first column is index and first row is colnames.
    matched_indies: If True, assumes X_df.columns == samples_df.index and X_df.index == guides_df.index
    """
    X_df = pd.read_csv(X_filename, index_col=0)
    if guides_filename:
        guides_df = pd.read_csv(guides_filename, index_col=0)
        if X_df.shape[0] != len(guides_df):
            raise ValueError(
                f"X has length {X_df.shape[0]} that does not match len(guides_df) {len(guides_df)}."
            )
    else:
        guides_df = None
    if samples_filename:
        samples_df = pd.read_csv(samples_filename, index_col=0)
        if X_df.shape[1] != len(samples_df):
            raise ValueError(
                f"X has width {X_df.shape[1]} that does not match len(samples_df) {len(samples_df)}."
            )
    else:
        samples_df = None

    if layers_filenames_dict:
        layers_dict = {}
        for key, df_path in layers_filenames_dict.items():
            layer_df = pd.read_csv(df_path, index_col=0)
    else:
        layers_dict = None

    if matched_indices:
        try:
            check_indices(X_df, guides_df.index, samples_df.index)
        except ValueError as e:
            raise ValueError("Error raised for X_df") from e
        if layers_filenames_dict:
            layers_dict = {}
            for key, df_path in layers_filenames_dict.items():
                try:
                    check_indices(layer_df, guides_df.index, samples_df.index)
                except ValueError as e:
                    raise ValueError("Error raised for layer_df") from e
                layers_dict[key] = layer_df
    else:
        if guides_df is not None:
            X_df = X_df.loc[guides_df.index, :]
            if layers_filenames_dict:
                for key, df_path in layers_filenames_dict.items():
                    layer_df = layer_df.loc[guides_df.index, :]
                    layers_dict[key] = layer_df
        if samples_df is not None:
            X_df = X_df.loc[:, samples_df.index]
            if layers_filenames_dict:
                for key, df_path in layers_filenames_dict.items():
                    layer_df = layer_df.loc[:, samples_df.index]
                    layers_dict[key] = layer_df
    if layers_dict:
        for key, df in layers_dict.items():
            layer_df = df.values

    return _Screen(
        X=X_df.values, guides=guides_df, samples=samples_df, layers=layers_dict
    )



import matplotlib
import os


def _set_matplotlib_rc_params():
    # os.system("rm ~/.cache/matplotlib -rf")

    font = {"size": 12}
    matplotlib.rc(font)
    matplotlib.rcParams["font.sans-serif"] = "Arial"
    matplotlib.rcParams["font.family"] = "sans-serif"



# _write_experiment_report_to_excel.py

__module_name__ = "_write_experiment_report_to_excel.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
import pandas as pd


# local imports #
# ------------- #
from ._check_fix_file_extension import _check_fix_file_extension
from ._write_screen_to_excel import _collect_screen_dfs


def _write_experiment_report_to_excel(
    screen,
    experiment_df_list,
    experiment_sheet_names,
    workbook_path="CRISPR.screen.experiment.summary.xlsx",
    silent=False,
    index=False,
):

    df_list, sheetnames = _collect_screen_dfs(screen)
    workbook_path = _check_fix_file_extension(workbook_path, ".xlsx", silent)

    for i in range(len(experiment_df_list)):
        df_list.append(experiment_df_list[i])
        sheetnames.append(experiment_sheet_names[i])

    assert len(df_list) == len(sheetnames), print(
        "Length of df_list must equal the length of sheet names passed."
    )

    with pd.ExcelWriter(workbook_path) as writer:

        if not silent:
            print("Writing to: {}\n".format(workbook_path))

        for n, df_ in enumerate(df_list):
            if not sheetnames:
                sheet = "Sheet_{}".format(n)
            else:
                sheet = sheetnames[n]
            if not silent:
                print("\tSheet {}:\t{}".format(str(int(n + 1)), sheet))
            df_.to_excel(writer, sheet, index=index)



import numpy as np

def _determine_overlap_3prime(
    fragment1_start, fragment1_stop, fragment2_start, fragment2_stop
):

    """
    Observe whether the start-site of fragment2 overlaps the 3' end of fragment1.

    Parameters:
    -----------
    fragment1_start

    fragment1_stop

    fragment2_start

    fragment2_stop

    Returns:
    --------
    overlap_3prime
        type: bool

    Notes:
    ------
    """

    return (fragment2_start < fragment1_stop) and (fragment2_stop > fragment1_stop)


def _determine_overlap_5prime(
    fragment1_start, fragment1_stop, fragment2_start, fragment2_stop
):

    """
    Observe whether the start-site of fragment2 overlaps the 5' end of fragment1.

    Parameters:
    -----------
    fragment1_start

    fragment1_stop

    fragment2_start

    fragment2_stop

    Returns:
    --------
    overlap_5prime
        type: bool

    Notes:
    ------
    """

    return (fragment2_start < fragment1_start) and (fragment2_stop > fragment1_start)


def _determine_partial_overlapping_fragments(df, silent, start_key="Start", end_key="End"):

    """

    Parameters:
    -----------
    df
        requires the following columns: ['start', 'stop']

    Returns:
    --------
    OverlapDict
    """

    OverlapDict = {}
    OverlapDict["5'"] = []
    OverlapDict["3'"] = []

    for fragment1 in range(len(df)):
        for fragment2 in range(len(df)):

            if _determine_overlap_3prime(
                df[start_key].iloc[fragment1],
                df[end_key].iloc[fragment1],
                df[start_key].iloc[fragment2],
                df[end_key].iloc[fragment2],
            ):
                OverlapDict["5'"].append([fragment1, fragment2])
                if not silent:
                    print(
                        "fragment {} overlaps the 5' end of fragment {}".format(
                            fragment1, fragment2
                        )
                    )

            if _determine_overlap_5prime(
                df[start_key].iloc[fragment1],
                df[end_key].iloc[fragment1],
                df[start_key].iloc[fragment2],
                df[end_key].iloc[fragment2],
            ):
                OverlapDict["3'"].append([fragment1, fragment2])
                if not silent:
                    print(
                        "fragment {} overlaps the 3' end of fragment {}".format(
                            fragment1, fragment2
                        )
                    )
    return OverlapDict


def _determine_total_overlap(
    fragment1_start, fragment1_stop, fragment2_start, fragment2_stop
):
    return (fragment1_start < fragment2_start) and (fragment1_stop > fragment2_stop)


def _determine_total_overlapping_fragments(df, silent=False, start_key="Start", end_key="End"):

    overlapped_fragments = []

    for fragment1 in range(len(df)):
        for fragment2 in range(len(df)):

            if _determine_total_overlap(
                df[start_key].iloc[fragment1],
                df[end_key].iloc[fragment1],
                df[start_key].iloc[fragment2],
                df[end_key].iloc[fragment2],
            ):
                overlapped_fragments.append(fragment1)
                if not silent:
                    print("fragment {} engulfs {}".format(fragment1, fragment2))
            elif _determine_total_overlap(
                df[start_key].iloc[fragment2],
                df[end_key].iloc[fragment2],
                df[start_key].iloc[fragment1],
                df[end_key].iloc[fragment1],
            ):
                overlapped_fragments.append(fragment2)
                if not silent:
                    print("fragment {} engulfs {}".format(fragment2, fragment1))

    return np.unique(overlapped_fragments)


def _drop_totally_overlapped_fragments(df, overlapped_fragments_list, start_key="Start", end_key="End"):

    """"""

    print("\nDropping the following exonic regions due to overlap:\n")
    for i in overlapped_fragments_list:
        print("\t{}-{}".format(df[start_key].iloc[i], df[end_key].iloc[i]))

    df_ = df.drop(overlapped_fragments_list)

    return df_.reset_index(drop=True)


def _identify_unique_fragments(df, start_key="Start", end_key="End"):

    df = df.reset_index(drop=True)
    df_ = (
        df.sort_values(start_key)
        .reset_index(drop=True)
        .drop_duplicates(subset=end_key, keep="first")
    )

    _df_ = (
        df_.sort_values(end_key)
        .reset_index(drop=True)
        .drop_duplicates(subset=start_key, keep="first")
    )

    return _df_.reset_index(drop=True)

class _OverlappingFragments:
    def __init__(self, df, silent=False):

        self.silent = silent
        if not self.silent:
            print("Removing duplicate rows")
        self.df = df.drop_duplicates()
        self.df = _identify_unique_fragments(df)
        self.df_raw = df
        self.OverlappingFragmentDict = {}

    def find_total(self):

        self.OverlappingFragmentDict["total"] = _determine_total_overlapping_fragments(
            self.df, self.silent
        )

        self.df = _drop_totally_overlapped_fragments(
            self.df, self.OverlappingFragmentDict["total"]
        )

    def find_equal_endpoint_fragments(self):

        self.OverlappingFragmentDict["longer_idx"] = _identify_unique_short_fragments(
            self.df
        )

        self.df = self.df.iloc[self.OverlappingFragmentDict["longer_idx"]]

    def find_partial(self):

        self.OverlappingFragmentDict[
            "partial"
        ] = _determine_partial_overlapping_fragments(self.df, self.silent)

    def find_all(self):

        self.OverlappingFragmentDict["total"] = _determine_total_overlapping_fragments(
            self.df, self.silent
        )

        self.df = _drop_totally_overlapped_fragments(
            self.df, self.OverlappingFragmentDict["total"]
        )
        self.OverlappingFragmentDict[
            "partial"
        ] = _determine_partial_overlapping_fragments(self.df, self.silent)


import seaborn as sns
import numpy as np
from scipy import stats
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib


def _annotate_pearsonr_foldchange(x, y, **kws):
    ignore_idx = np.isnan(x) | np.isnan(y)
    x = x[~ignore_idx]
    y = y[~ignore_idx]
    r, _ = stats.pearsonr(x, y)
    ax = plt.gca()

    # count how many annotations are already present
    n = len([c for c in ax.get_children() if isinstance(c, matplotlib.text.Annotation)])
    pos = (0.1, 0.9 - 0.1 * n)

    ax.annotate("r={:.2f}".format(r), xy=pos, xycoords=ax.transAxes)


def plot_rep_consistency_lfc(
    screen,
    cond1,
    cond2,
    hue=None,
    lognorm_counts_key="lognorm_counts",
    rep_condit="replicate",
    compare_col="sort",
    out_guides_suffix="lfc",
    keep_rep_lfc=False,
    title="Replicate consistency of LFC",
    return_plot=True,
    **kws
):

    if not hue is None and not hue in screen.guides.columns:
        raise KeyError("{} not in guide features.".format(hue))

    lfcs_df_columns = [
        s + ".{}_{}.{}".format(cond1, cond2, out_guides_suffix)
        for s in screen.samples[rep_condit].unique()
    ]
    lfcs_df = screen.log_fold_change_reps(
        cond1,
        cond2,
        lognorm_counts_key=lognorm_counts_key,
        rep_condit=rep_condit,
        compare_condit=compare_condit,
        out_guides_suffix=out_guides_suffix,
        keep_result=keep_rep_lfc,
    )

    if not hue is None:
        lfcs_df = pd.concat((screen.guides[[hue]], lfcs_df), axis=1)
    try:
        pairplot_obj = sns.pairplot(
            lfcs_df, kind="scatter", hue=hue, corner=True, **kws
        )
    except:
        return lfcs_df
    pairplot_obj.map_lower(_annotate_pearsonr_foldchange)
    # pairplot_obj.fig.suptitle(title)
    # pairplot_obj.fig.tight_layout()
    if return_plot:
        return pairplot_obj




import numpy as np

from ._calculate_log_fold_change import _calculate_enrichment_pvalues
from ._calculate_log_fold_change import _calculate_baseline_subtraction
from ._calculate_log_fold_change import _calculate_delta_logfoldchange

#from ...._plotting._funcs._plot_guide_enrichment import _plot_guide_enrichment

class LogFoldChangeModule:
    def __init__(self, counts_df):

        """
        Calculate the log fold change between two conditions.

        Parameters:
        -----------
        counts_df
            example: screen.layers["lognorm_counts"]
            type: pandas.core.frame.DataFrame

        condit_1
            Regex string by which to identify condition columns in counts_df that match condit_1.
            type: str

        condit_2
            Regex string by which to identify condition columns in counts_df that match condit_2.
            type: str

        control
            Regex string by which to identify condition columns in counts_df that match control condition.
            type: str


        Returns:
        --------

        Notes:
        ------
        (1) Assumes that two conditions are to be compared, originating from a common initial condition.
        (2) Assumes log-transformed values.
        """

        self.counts_df = counts_df

    def filter_guides_by_experiments(self, guides, targets):

        """
        Remove guides that do not target the loci of interest.

        Control guides may be shared across experiments.

        Parameters:
        -----------
        counts_df

        screen.guides

        list(str)

        Notes:
        ------
        counts_df and guides should have the same length with (initially) matching indices.
        """

        self.counts_df = self.counts_df.iloc[
            guides["target"].loc[guides["target"].isin(targets)].index.astype(int)
        ]

    def isolate_conditions(self, condit_1, condit_2, control):

        self.condit_1_ = self.counts_df.filter(regex=condit_1)
        self.condit_2_ = self.counts_df.filter(regex=condit_2)
        self.control_ = self.counts_df.filter(regex=control)

    def calculate_logfoldchange(self):

        """
        Three steps:
        ------------
        (1) calculate baseline.
        (2) calculate baseline-LFC.
        (3) Calculate a p-value for the logfoldchange of each guide between conditions.

        """

        (
            self.condit_1_vs_control,
            self.condit_2_vs_control,
        ) = _calculate_baseline_subtraction(
            self.condit_1_, self.condit_2_, self.control_
        )

        self.lfc_mean, self.lfc_stdev = _calculate_delta_logfoldchange(
            self.condit_1_vs_control, self.condit_2_vs_control
        )

        self.p_values = _calculate_enrichment_pvalues(
            self.condit_1_vs_control, self.condit_2_vs_control
        )
        
        
def _annotate_guides_with_log_fold_change(screen, lfc_module):

    """"""

    experiment_guide_df = screen.guides.iloc[lfc_module.counts_df.index]
    experiment_guide_df["lfc.mean"] = lfc_module.lfc_mean
    try:
        experiment_guide_df["lfc.stdev"] = lfc_module.lfc_stdev
        experiment_guide_df["lfc.p_value"] = lfc_module.p_values
        experiment_guide_df["-log10(pval)"] = -np.log10(experiment_guide_df["lfc.p_value"])
    except:
        pass

    return experiment_guide_df
        
    
def _log_fold_change_analysis(
    screen,
    condit_1,
    condit_2,
    control,
    targets,
    layer="lognorm_counts",
    plot=True,
    plot_title=None,
    return_fig=False,
    gene=False,
    gtf_path=None,
):

    """"""

    lfc = LogFoldChangeModule(screen.layers[layer])
    lfc.filter_guides_by_experiments(screen.guides, targets)
    lfc.isolate_conditions(condit_1, condit_2, control)
    lfc.calculate_logfoldchange()
    
    experiment_guide_df = _annotate_guides_with_log_fold_change(screen, lfc)
    
    if plot:
        fig = _plot_guide_enrichment(experiment_guide_df, title=plot_title, return_fig=return_fig, gene=gene, gtf_path=gtf_path)
        if return_fig:
            return experiment_guide_df, fig
    else:
        return experiment_guide_df



import numpy as np

def _annotate_protospacer(guide_df, protospacer_length=20):

    if np.all(guide_df["barcode"].str.len().unique() == protospacer_length):
        guide_df["protospacer"] = guide_df["barcode"]

    return guide_df




from ....._utilities._funcs._SequenceManipulation import _SequenceManipulation
from ....._utilities._funcs._fetch_chromosome_sequence import _fetch_chromosome_sequence

def _load_relevant_chromosomes(gene_df, ref_seq_path, silent=False):

    """
    Load unique chromosomes found in the gene dict into memory


    Notes:
    ------
    (1) Doing this up front and once prevents the redundant fetching of chromosome sequences, which can be slow.
    (2) The produced dict can be deleted (and the corresponding memory freed) upon substringing the chromosome.
    """

    ChromDict = {}

    for chrom in gene_df.Chromosome.unique():
        if not silent:
            print("loading {}...".format(chrom))
        ChromDict[chrom] = _fetch_chromosome_sequence(ref_seq_path, chrom)
    return ChromDict

def _add_context_sequence_to_gene_dict(gene_dict, gene_df, ref_seq_path):

    ChromDict = _load_relevant_chromosomes(gene_df, ref_seq_path)

    for gene, values in gene_dict.items():
        gene_dict[gene]["seq"] = {}
        gene_dict[gene]["chrom.length"] = len(ChromDict[values["Chromosome"]])
        gene_start = int(values["Start"] * 1e6)
        gene_end = int(values["End"] * 1e6)
        sequence = _SequenceManipulation(
            ChromDict[values["Chromosome"]][gene_start:gene_end]
        )

        gene_dict[gene]["seq"]["+"] = sequence.sequence
        gene_dict[gene]["seq"]["-"] = sequence.reverse_complement()

    print("\nDeleteting whole chromosome sequences from memory...")
    del ChromDict

    return gene_dict


# _update_dict.py

__module_name__ = "_update_dict.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


def _update_dict(self, **kwargs):
    
    """Update a dictionary given some keyword argument pair."""

    for key, value in kwargs.items():
        self.X[key] = value

    return self



# import plotly.express as px
from tqdm import tqdm_notebook as tqdm

#from ..._utilities._funcs._get_gene_exons import _get_gene_exons

from ._plot_gene_exon import _plot_gene_exon

def _plot_guide_outline(fig, start, end, y):

    """"""

    fig.add_shape(
        x0=start,
        x1=end,
        y0=0,
        y1=y,
        line_width=2,
    )


    
def _plot_guide_fill(fig, start, end, y, fillcolor="navy", opacity=0.1):

    """"""

    fig.add_shape(
        x0=start,
        x1=end,
        y0=0,
        y1=y,
        line_width=0,
        fillcolor=fillcolor,
        opacity=opacity,
    )



def _plot_single_guide(fig, start, end, y, fillcolor="navy", opacity=0.1):

    _plot_guide_outline(fig, start, end, y)
    _plot_guide_fill(fig, start, end, y, fillcolor="navy", opacity=0.1)


def _format_plot(fig, 
                 experiment_guide_df, 
                 y_title="\u0394Log\u2082(Fold Change)", 
                 gridcolor="mistyrose"):

    """"""
    fig.update_layout(
        {
            "plot_bgcolor": "rgba(0, 0, 0, 0)",
            "paper_bgcolor": "rgba(0, 0, 0, 0)",
        }
    )
    fig.add_hline(y=0.0, line_width=0.5, line_dash="dash", line_color="red")
    fig.update_layout(font_family="Avenir", title_font_family="sans-serif")
    fig.update_yaxes(title=y_title)
    fig.update_xaxes(
        title="Genomic Coordinates: {}".format(experiment_guide_df.Chromosome.unique()[0])
    )
    fig.update_xaxes(showgrid=True, gridwidth=1, gridcolor=gridcolor)
    fig.update_yaxes(showgrid=True, gridwidth=1, gridcolor=gridcolor)


def _plot_guide_enrichment(experiment_guide_df, title=None, return_fig=False, exon_color='red', gene=False, gtf_path=None):

    """

    Minimal func for now. Lacks several helpful / usability things and is very stiff.

    """

    fig = px.scatter(
        experiment_guide_df,
        title=title,
        x="center",
        y="lfc.mean",
    )

    _format_plot(fig, experiment_guide_df)

    for i, row in tqdm(experiment_guide_df.iterrows()):
        _plot_single_guide(fig, row["Start"], row["End"], row["lfc.mean"])
    
    if gene:
        if type(gene) == str:
            gene = [gene]
        for i in range(len(gene)):
            gene_df = _get_gene_exons(gene[i], experiment_guide_df.Chromosome.unique()[0], gtf_path)
            _plot_gene_exon(fig, gene[i], gene_df, lfc=experiment_guide_df['lfc.mean'], exon_color=exon_color)
    fig.show()

    if return_fig:
        return fig



# _fetch_chromosome_sequence.py

from Bio import SeqIO

def _fetch_chromosome_sequence(ref_seq_path, query_chr):
    
    """
    Isolate a specific chromosomal sequence (forward annotation) from a reference genome. 
    
    ref_seq_path
        Path to reference genome sequence.
        type: str
        
    query_chr
        example: 'chr2'
        type: str
        
    silent
        default: False
        default: boolean
    """
    
    for record in SeqIO.parse(ref_seq_path, "fasta"):
        if record.description.split()[0] == query_chr:
            return str(record.seq)




#import vintools as v

def _create_gene_dict(genes, chrom, start, stop):

    """"""
    NotImplemented
    # GeneDict = v.ut.EmptyDict(genes)

    # for n, gene in enumerate(GeneDict.keys()):
    #     GeneDict[gene]["Chromosome"] = chrom[n]
    #     GeneDict[gene]["Start"] = start[n]
    #     GeneDict[gene]["End"] = stop[n]

    # return GeneDict



import pandas as pd

def _gene_from_gtf(gtf_file, chromosome, gene):

    """"""

    GeneDict = {}

    n = 0
    for i, line in enumerate(gtf_file.readlines()):
        line = line.strip("\n").split("\t")
        if line[0] == chromosome:
            line_ = line[8].split(" ")
            if line_[10] == "gene_name":
                if line_[11].startswith('"{}'.format(gene)):
                    GeneDict[n] = line
                    n += 1
    df = pd.DataFrame.from_dict(GeneDict, orient="index")

    return df


def _get_feature(df, feature):
    return df.loc[df[2] == feature]


def _get_gene_body_bounds(gene_df):

    gene_body_min = gene_df[3].astype(int).min()
    gene_body_max = gene_df[4].astype(int).max()

    return gene_body_min, gene_body_max


# _ScreenModule.py
__module_name__ = "_ScreenModule.py"
__author__ = ", ".join(["Michael E. Vinyard", "Jayoung Kim Ryu"])
__email__ = ", ".join(["vinyard@g.harvard.edu", "jayoung_ryu@g.harvard.edu"])

import warnings
from typing import Union, List

import anndata as ad
import numpy as np
import pandas as pd
from anndata import AnnData

from .._arithmetic._funcs._log_fold_change import _log_fold_change
from .._normalization._funcs._read_count_norm import _log_normalize_read_count
from .._readwrite._funcs._read_screen_from_PoolQ import _read_screen_from_PoolQ
from .._readwrite._funcs._write_screen_to_csv import _write_screen_to_csv
from .._readwrite._funcs._write_screen_to_excel import _write_screen_to_excel
from .._utilities._funcs._update_dict import _update_dict
from ._supporting_functions._guides._GuideAnnotationModule import _annotate_sgRNAs
from ._supporting_functions._print_screen_object import _print_screen_object


class _Screen(AnnData):
    def __init__(self, X=None, guides=None, samples=None, *args, **kwargs):
        super().__init__(X=X, obs=guides, var=samples, *args, **kwargs)

    @classmethod
    def from_adata(cls, adata: ad.AnnData):
        return cls(
            (adata.X),
            guides=(adata.obs),
            samples=(adata.var),
            obsm=adata.obsm,
            obsp=adata.obsp,
            uns=(adata.uns),
            layers=(adata.layers),
        )

    @property
    def guides(self):
        return self.obs

    @guides.setter
    def guides(self, new_val: pd.DataFrame):
        self.obs = new_val

    @property
    def samples(self):
        return self.var

    @samples.setter
    def samples(self, new_val: pd.DataFrame):
        self.var = new_val

    @property
    def samples_m(self):
        return self.varm

    @samples_m.setter
    def samples_m(self, new_val):
        self.varm = new_val

    @property
    def samples_p(self):
        return self.varp

    @samples_p.setter
    def samples_p(self, new_val):
        self.varp = new_val

    def __repr__(self) -> str:
        return _print_screen_object(self)[2]

    def __add__(self, other):
        if all(self.guides.index == other.guides.index) and all(
            self.samples.index == other.samples.index
        ):
            return _Screen(self.X + other.X, self.guides.copy(), self.samples.copy())
        else:
            raise ValueError("Guides/sample description mismatch")

    def __getitem__(self, index):
        adata = super().__getitem__(index)
        return _Screen.from_adata(adata)

    def read_PoolQ(self, path, metadata=False, merge_metadata_on="Condition"):
        """Read poolQ."""
        self._PoolQ_outpath = path
        self._PoolQScreenDict = _read_screen_from_PoolQ(self._PoolQ_outpath)

        for key, value in _update_dict(self._PoolQScreenDict).items():
            self.__setattr__(key, value)

        if metadata:
            self.samples = self.samples.merge(
                pd.read_csv(metadata), on=merge_metadata_on
            )

        _print_screen_object(self)

    def annotate_guides(
        self, genes, chrom, start, stop, annotations, DirectPairDict, ref_seq_path
    ):
        """
        Annotate sgRNA table.

        """
        self.guides = _annotate_sgRNAs(
            self.guides,
            genes,
            chrom,
            start,
            stop,
            annotations,
            DirectPairDict,
            ref_seq_path,
        )

    def log_norm(self, output_layer="lognorm_counts", read_count_layer=None, pseudocount = 1):
        if read_count_layer is None:
            self.layers[output_layer] = _log_normalize_read_count(self.X, pseudocount)
        else:
            output_layer = f"lognorm_{read_count_layer}"
            self.layers[output_layer] = _log_normalize_read_count(
                self.layers[read_count_layer], pseudocount
            )

    # TBD: mask ones with too low raw counts.
    def log_fold_change(
        self,
        sample1,
        sample2,
        lognorm_counts_key="lognorm_counts",
        name=False,
        pseudocount: int = 1,
        out_guides_suffix="lfc",
        return_result=False,
    ):
        """
        General module to calculate LFC across experimental conditions.
        """
        if lognorm_counts_key == "lognorm_counts":
            self.log_norm(pseudocount=pseudocount)
        else:
            if "lognorm_" not in lognorm_counts_key:
                raise ValueError(f"{lognorm_counts_key} is not a lognorm layer- feed in 'lognorm_`layer_key`' as lognorm_counts_key.")
            read_count_layer_key = lognorm_counts_key.split("lognorm_")[-1]
            if read_count_layer_key not in self.layers:
                raise ValueError(f"{read_count_layer_key} not in .layers - feed in 'lognorm_`layer_key`' as lognorm_counts_key.")
            self.log_norm(output_layer=lognorm_counts_key, read_count_layer=read_count_layer_key, pseudocount=pseudocount)
        sample1_idx = np.where(sample1 == self.samples.index)[0]
        sample2_idx = np.where(sample2 == self.samples.index)[0]
        if len(sample1_idx) != 1 or len(sample2_idx) != 1:
            if len(sample1_idx) == 0:
                print(f"No sample named {sample1} in Screen object.")
            else:
                print(f"Duplicate sample name {sample1} in Screen object")
            if len(sample2_idx) == 0:
                print(f"No sample named {sample2} in Screen object.")
            else:
                print(f"Duplicate sample name {sample2} in Screen object")
            raise ValueError("")

        try:
            lfc = _log_fold_change(
                self.layers[lognorm_counts_key], sample1_idx, sample2_idx
            )
            if return_result:
                return lfc
            else:
                self.guides[f"{sample1}_{sample2}.{out_guides_suffix}"] = lfc
        except Exception:  # TBD: what error?
            print("Calculating LFC against two previously calculated LFC values...")
            dlfc = _log_fold_change(self.guides, sample1, sample2)

            if not name:
                name = f'{sample1.strip(".lfc")}_{sample2.strip(".lfc")}.d{out_guides_suffix}'

            if return_result:
                return dlfc
            else:
                self.guides[name] = dlfc

    def log_fold_change_reps(
        self,
        cond1,
        cond2,
        lognorm_counts_key="lognorm_counts",
        rep_col: Union[str, List[str]] = "replicate",
        compare_col="sort",
        out_guides_suffix="lfc",
        pseudocount=1,
        keep_result=False,
        ignore_missing=False,
    ):
        """Get gRNA abundance LFC across conditions for each replicate.
        ignore_missing: If True, does not raise Error when one of the conditions is missing for a replicate.
        """
        if isinstance(rep_col, str) and rep_col not in self.samples.columns:
            raise ValueError(f"{rep_col} not in samples features")
        elif isinstance(rep_col, list):
            for rc in rep_col:
                if rc not in self.samples.columns:
                    raise ValueError(f"{rc} not in samples features")
        if compare_col not in self.samples.columns:
            raise ValueError(f"{compare_col} not in samples features")
        if (
            cond1 not in self.samples[compare_col].tolist()
            or cond2 not in self.samples[compare_col].tolist()
        ):
            raise ValueError(
                f"{cond1} or {cond2} not in sample conditions {self.samples[compare_col]}"
            )

        lfcs = []

        if isinstance(rep_col, str):
            lfc_rep_cols = self.samples[rep_col].unique().tolist()
            unique_reps = self.samples[rep_col].unique()
        else:
            unique_reps = [
                list(t)
                for t in self.samples[rep_col].drop_duplicates().to_records(index=False)
            ]
            lfc_rep_cols = [
                ".".join(list(map(str, rep_list))) for rep_list in unique_reps
            ]
        for rep in unique_reps:
            if isinstance(rep_col, str):
                cond1_idx = np.where(
                    (self.samples[rep_col] == rep)
                    & (self.samples[compare_col] == cond1)
                )[0]
                cond2_idx = np.where(
                    (self.samples[rep_col] == rep)
                    & (self.samples[compare_col] == cond2)
                )[0]
            elif isinstance(rep_col, list):
                cond1_idx = np.where(
                    (self.samples[rep_col] == rep).all(axis=1)
                    & (self.samples[compare_col] == cond1)
                )[0]
                cond2_idx = np.where(
                    (self.samples[rep_col] == rep).all(axis=1)
                    & (self.samples[compare_col] == cond2)
                )[0]
            if len(cond1_idx) != 1 or len(cond2_idx) != 1:
                if not ignore_missing:
                    raise ValueError(
                        f"Conditions are not unique for each replicates ({rep_col} == {rep}) to be aggregated:\n{cond1}:{self.samples[cond1_idx]} or {cond2}:{self.samples[cond2_idx]}\n {self.samples}"
                    )
                else:
                    lfc_rep_cols.pop(rep)
                    continue

            lfcs.append(
                self.log_fold_change(
                    self.samples.index[cond1_idx].tolist()[0],
                    self.samples.index[cond2_idx].tolist()[0],
                    lognorm_counts_key=lognorm_counts_key,
                    pseudocount=pseudocount,
                    return_result=True,
                )
            )

        lfcs_array = np.concatenate(lfcs, axis=1)
        lfcs_df_columns = [
            f"{s}.{cond1}_{cond2}.{out_guides_suffix}" for s in lfc_rep_cols
        ]
        lfcs_df = pd.DataFrame(
            lfcs_array, index=self.guides.index, columns=lfcs_df_columns
        )

        if keep_result:
            self.guides[lfcs_df_columns] = lfcs_df
        return lfcs_df

    # TODO: add guides metadata on how aggregates are calcualted?
    def log_fold_change_agg(
        self,
        cond1,
        cond2,
        lognorm_counts_key="lognorm_counts",
        agg_col="replicate",
        compare_col="condition",
        out_guides_suffix="lfc",
        agg_fn="median",
        pseudocount=1,
        name=None,
        return_result=False,
        keep_per_replicate=False,
    ):
        lfcs_df = self.log_fold_change_reps(
            cond1,
            cond2,
            lognorm_counts_key=lognorm_counts_key,
            rep_col=agg_col,
            compare_col=compare_col,
            out_guides_suffix=out_guides_suffix,
            pseudocount=pseudocount,
            keep_result=keep_per_replicate,
        )

        if agg_fn == "mean":
            lfcs_agg = lfcs_df.apply(np.mean, axis=1)
        elif agg_fn == "median":
            lfcs_agg = lfcs_df.apply(np.median, axis=1)
        elif agg_fn == "sd":
            lfcs_agg = lfcs_df.apply(np.std, axis=1)
        else:
            raise ValueError(
                "Only 'mean', 'median', and 'sd' are supported for aggregating LFCs."
            )

        if return_result:
            return lfcs_agg
        if name is None:
            self.guides[f"{cond1}_{cond2}.{out_guides_suffix}.{agg_fn}"] = lfcs_agg
        else:
            self.guides[name] = lfcs_agg

    def fold_change(
        self,
        cond1: str,
        cond2: str,
        lognorm_counts_key: str = "lognorm_counts",
        return_result: bool = False,
    ) -> Union[None, pd.Series]:
        """Calculate log fold change (cond1/cond2) of normalized guide abundances."""

        log_fold_change = _log_fold_change(
            self.layers[lognorm_counts_key], cond1, cond2
        )
        if return_result:
            return log_fold_change
        self.guides[f"{cond1}_{cond2}.fc"] = log_fold_change

    def to_Excel(
        self,
        workbook_path: str = "CRISPR_screen.workbook.xlsx",
        index: bool = False,
        silent: bool = False,
        include_uns: bool = False,
    ) -> None:
        """Write components of Screen class to an Excel workbook.

        Args
        ---
        workbook_path: Prevent printing outpaths / details of the created workbook.
        index: If True, include an index in the workbook sheet.
        silent: If True, prevent printing outpaths / details of the created workbook.

        Notes:
        ------
        (1) Will likely need to be updated once we fully transition over to AnnData-like class.
        """

        _write_screen_to_excel(
            self,
            workbook_path,
            index,
            silent,
            include_uns,
        )

    def to_csv(self, out_path="CRISPR_screen"):
        """

        Write .csv files for each part of the screen. will eventually be replaced by something more native to AnnData.

        """

        _write_screen_to_csv(self, out_path)

    def to_mageck_input(
        self,
        out_path=None,
        count_layer=None,
        sgrna_column=None,
        target_column="target_id",
        sample_prefix="",
    ):
        """Formats screen object into mageck input.

        If screen.guides[target_column] is None or np.isnan, remove that guide
        """
        if count_layer is None:
            count_matrix = self.X
        else:
            try:
                count_matrix = self.layers[count_layer]
            except KeyError as exc:
                raise KeyError(
                    f"Layer {count_layer} doesn't exist in Screen object with layers {self.layers.keys()}"
                ) from exc
        mageck_input_df = (
            pd.DataFrame(
                count_matrix,
                columns=sample_prefix + self.samples.index,
                index=self.guides.index,
            )
            .fillna(0)
            .astype(int)
        )
        if sgrna_column is None:
            mageck_input_df.insert(0, "sgRNA", self.guides.index.tolist())
        elif sgrna_column in self.guides.columns:
            mageck_input_df.insert(0, "sgRNA", self.guides[sgrna_column])
        elif self.guides.index.name == sgrna_column:
            mageck_input_df.insert(0, "sgRNA", self.guides.index.tolist())
        else:
            raise ValueError(f"{sgrna_column} not found in Screen.guides.")
        mageck_input_df["sgRNA"] = mageck_input_df["sgRNA"].map(
            lambda s: s.replace(" ", "_")
        )
        mageck_input_df.insert(1, "gene", self.guides[target_column].astype(str))
        mageck_input_df = mageck_input_df.loc[
            (mageck_input_df.gene.map(lambda o: not pd.isnull(o)))
            & mageck_input_df.gene.map(bool),
            :,
        ]
        if out_path is None:
            return mageck_input_df
        else:
            mageck_input_df.to_csv(out_path, sep="\t", index=False)

    def write(self, out_path):
        """
        Write .h5ad
        """
        super().write(out_path)


def read_h5ad(filename):
    adata = ad.read_h5ad(filename)
    return _Screen.from_adata(adata)


def concat(screens, *args, **kwargs):
    adata = ad.concat(screens, *args, **kwargs)

    return _Screen.from_adata(adata)


def read_csv(X_path=None, guides_path=None, samples_path=None, sep=",", **kwargs):
    if X_path is not None:
        X_df = pd.read_csv(
            X_path,
            delimiter=sep,
            header=0,
            index_col=0,
        )
        X = X_df.values
    else:
        X = None
    guide_df = (
        pd.read_csv(guides_path, sep=sep, **kwargs) if guides_path is not None else None
    )
    samples_df = None if samples_path is None else pd.read_csv(samples_path, sep=sep)
    return _Screen(X=X, guides=guide_df, samples=samples_df)



def _annotate_homopolymer(df):
    
    """Annotate any homopolymer sequences in sgRNAs"""
    
    df['polyT'] = df.protospacer.str.contains("TTTT")
    df['polyG'] = df.protospacer.str.contains("GGGG")
    df['polyC'] = df.protospacer.str.contains("CCCC")
    df['polyA'] = df.protospacer.str.contains("AAAA")
    
    return df

def _get_GC_content(df):
    
    """Get the GC content of sgRNAs"""
    
    g_count = df.protospacer.str.count("G")
    c_count = df.protospacer.str.count("C")
    
    df['GC_content'] = (g_count + c_count)*100/20
    
    return df

def _annotate_guide_biochemistry(df):
    
    """Adds information to the guide df about GC content and polyT content"""
    
    df = _annotate_homopolymer(df)
    df = _get_GC_content(df)
    
    return df


import anndata as ad
import numpy as np
import pandas as pd
import perturb_tools as pt
import pytest


@pytest.fixture
def screen():
    return pt.Screen(
        X=np.random.rand(3, 4),
        guides=pd.DataFrame(
            np.random.rand(3, 1),
            columns=[
                "a",
            ],
            index=["g1", "g2", "g3"],
        ),
        samples=pd.DataFrame(index=["c1", "c2", "c3", "c4"]),
    )


@pytest.fixture
def adata():
    return ad.AnnData(
        X=np.random.rand(3, 4),
        obs=pd.DataFrame(
            np.random.rand(3, 1),
            columns=[
                "a",
            ],
            index=["g1", "g2", "g3"],
        ),
        var=pd.DataFrame(index=["c1", "c2", "c3", "c4"]),
    )


class TestClass:
    def test_read(self, screen):
        screen.write("tmp")
        pt.read_h5ad("tmp")

    def test_subset(self, screen):
        screen_sub = screen[["g1", "g2"], :]
        assert screen_sub.shape == (2, 4)
        assert len(screen_sub.obs) == 2

    def test_select(self, screen):
        screen_sel = screen[:, ["c1", "c3"]]
        assert screen_sel.shape == (3, 2)
        assert len(screen_sel.var) == 2

    def test_from_adata(self, adata, screen):
        screen = pt.Screen.from_adata(adata)
        assert (screen.guides.index == adata.obs.index).all()
        assert (screen.samples.index == adata.var.index).all()




import numpy as np

def _add_region_labels(guide_df):
    
    """"""
    
    regions = np.array([])
    
    for n, region in enumerate(guide_df.groupby("region_left")):
        regions = np.append(regions, np.full(len(region[1]), n))
        
    guide_df['region'] = regions
    
    return guide_df



# _flexible_mkdir.py
__module_name__ = "_flexible_mkdir.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
#import licorice
import os


# local imports #
# ------------- #


def _flexible_mkdir(path, verbose):

    if os.path.exists(path):
        pass
    elif path == '':
        pass
    else:
        os.mkdir(path)
        # if verbose:
        #     msg = licorice.font_format("Directory created", ["BOLD", "CYAN"])
        #     print("{}: {}".format(msg, path))


def _flexible_multilevel_mkdir(path, verbose=True):

    """
    Create a directory or ignore if already present. Can create multiple levels of directories.
    Parameters:
    -----------
    path
    Returns:
    --------
    None
        os.mkdir(path) (or nothing)
    Notes:
    ------
    (1) supports multi-level directory creation
    """

    parsed_path = []

    for n, directory in enumerate(path.split("/")):
                
        parsed_path.append(directory)
        if len(parsed_path) == (len(path.split("/")) - 1):
            _flexible_mkdir("/".join(parsed_path), verbose)
        else:
            _flexible_mkdir("/".join(parsed_path), verbose=False)



from ..._arithmetic._funcs._SequenceManipulation import _SequenceManipulation
from ..._utilities._funcs._scan_sequence_for_motif import _scan_sequence_for_motif

import numpy as np

def _scan_for_BsmbI(df, filter_BsmbI=True):

    """"""

    forward_motif_search = "TCTC"
    reverse_motif_search = "GAGAC"

    BsmbI_flag = []
    count = 0
    for a, guide in enumerate(df.guide_context):
        
        
        guide_seq = _SequenceManipulation(guide)
        rev_guide = guide_seq.reverse_complement()

        a = _scan_sequence_for_motif(guide, forward_motif_search, negative_report=True)
        b = _scan_sequence_for_motif(guide, reverse_motif_search, negative_report=True)
        c = _scan_sequence_for_motif(
            rev_guide, forward_motif_search, negative_report=True
        )
        d = _scan_sequence_for_motif(
            rev_guide, reverse_motif_search, negative_report=True
        )

        x = np.array([a, b, c, d])

        if np.any(x) == True:
            count += 1
            BsmbI_flag.append(True)
        else:
            BsmbI_flag.append(False)
    print(count, "sgRNAs were found to contain BsmbI sites and were flagged.")

    df["BsmbI_flag"] = BsmbI_flag
    
    if filter_BsmbI:
        print("\nFiltering BsmbI-containing sgRNAs...\n")
        BsmbI_guide_df = df.loc[df.BsmbI_flag == True]
        df = df.loc[df.BsmbI_flag == False]
        return df, BsmbI_guide_df

    return df, None


import numpy as np


def _read_count_normalize(X, pseudocount: int = 1):
    """Read depth normalization by sample. Assumes samples are columns and guides are rows."""

    return (X / np.nansum(X+pseudocount, axis=0)) * 1e6


def _log_transform_read_count(X, pseudocount: int = 1):
    """"""
    return np.log2(X + pseudocount)


def _log_normalize_read_count(X, pseudocount:int = 1):
    """Following the protocol written clearly, here:
    https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0170445#sec002
    (see Methods).
    """

    X_read_norm = _read_count_normalize(X, pseudocount)
    X_log_read_norm = _log_transform_read_count(X_read_norm)

    return X_log_read_norm



# _external_tools (.tl) __init__.py


# __init__.py
__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


#from pool_sharq import poolq



import numpy as np
import matplotlib.pyplot as plt

# https://stackoverflow.com/questions/39512260/calculating-gini-coefficient-in-python-numpy
def G(v):
    bins = np.linspace(0.0, 1.0, 50)
    total = float(np.sum(v))
    yvals = []
    for b in bins:
        bin_vals = v[v <= np.quantile(v, b)]
        bin_fraction = np.sum(bin_vals) / total
        yvals.append(bin_fraction)
    # perfect equality area
    pe_area = np.trapz(bins, x=bins)
    # lorenz area
    lorenz_area = np.trapz(yvals, x=bins)
    gini_val = (pe_area - lorenz_area) / float(pe_area)
    return bins, yvals, gini_val


def plot_gini(adata, ax):
    for c in adata.samples.index:
        bins, result, gini_val = G(adata[:, c].X)
        ax.plot(bins, result, label="{} ({:.3f})".format(c, gini_val))
    ax.plot((0, 1), (0, 1), "--", label="perfect eq.")
    ax.set_xlabel("Cumulative fraction of guides \nfrom lowest to highest read counts")
    ax.set_ylabel("Cumulative fraction of read counts")
    ax.set_title("GINI index")
    ax.set_legend(bbox_to_anchor=(1.02, 1))


def plot_count(adata, ax):
    ax.figure()
    bins = np.linspace(0, adata.X.max(), 100)
    for c in adata.samples.index:
        ax.hist(
            adata[:, c].X,
            bins,
            alpha=0.5,
            label="{} (median={})".format(c, np.nanmedian(adata[:, c].X)),
        )
    ax.set_xlabel("Cumulative fraction of guides \nfrom lowest to highest read counts")
    ax.set_ylabel("Cumulative fraction of read counts")
    ax.set_title("GINI index")
    ax.set_legend(bbox_to_anchor=(1.02, 1))




def _scan_sequence_for_motif(seq, motif, negative_report=False):

    """Return location of motif if found"""
    
    if motif in seq:
        return seq.find(motif)
    if negative_report == True:
        return False


import numpy as np
import numpy.ma as ma
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

linestyles = ["solid", "dotted", "dashed", "dashdot"]


def spearman_keepdim(X):
    # print(scipy.stats.spearmanr(X, nan_policy="omit"))
    # c = scipy.stats.spearmanr(X, nan_policy="omit")[0]
    df = pd.DataFrame(X)
    c = df.corr(method="spearman").values
    return c


corr_func_dict = {
    "Pearson": lambda X: ma.corrcoef(X.T),
    "Spearman": lambda X: spearman_keepdim(X),
}


def set_count_stats(screen):
    screen.samples["median_X"] = np.nanmedian(screen.X, axis=0)
    screen.samples["median_X_bcmatch"] = np.nanmedian(
        screen.layers["X_bcmatch"], axis=0
    )
    if "edits" in screen.layers.keys():
        screen.samples["median_edit"] = np.nanmedian(screen.layers["edits"], axis=0)
        screen.get_guide_edit_rate()
        screen.samples["median_edit_rate"] = np.nanmedian(
            screen.layers["edit_rate"], axis=0
        )


def set_sample_correlation_guides(screen, guide_idx, prefix="", method="Pearson"):
    corr_func = corr_func_dict[method]
    screen_subset = screen[guide_idx, :].copy()
    if prefix != "":
        prefix = f"{prefix}_"
    c = corr_func(ma.masked_invalid(screen_subset.X))
    if np.isnan(c).all():
        print(
            f"Failed to calculate {method} correlation. Check if your matrix is constant or have no valid values."
        )
        return
    screen.varm[f"{prefix}corr_X"] = c

    # screen.samples[f"{prefix}mean_corr_X"] = screen.varm[f"{prefix}corr_X"].mean(0)

    screen.samples[f"{prefix}median_corr_X"] = np.nanmedian(
        screen.varm[f"{prefix}corr_X"], axis=0
    )


def set_sample_correlation(screen, method="Pearson"):
    set_sample_correlation_guides(screen, screen.guides.index, prefix="", method=method)


def plot_correlation(screen, method="Pearson"):
    """
    arguments
    -- method: ["Pearson", "Spearman"]
    """
    set_sample_correlation(screen, method)
    n_corrs = len(screen.varm.keys())

    def plot_heatmap(k, cor, ax):
        sns.heatmap(
            cor,
            ax=ax,
            xticklabels=screen.samples.index,
            yticklabels=screen.samples.index,
        )
        ax.set_title(k)
        ax.set_aspect(1)

    fig, ax = plt.subplots(n_corrs, 1, figsize=(10, n_corrs * 10))
    for i, (k, cor) in enumerate(screen.varm.items()):
        if n_corrs > 1:
            plot_heatmap(k, cor, ax[i])
        else:
            plot_heatmap(k, cor, ax)
            break
    return ax


def plot_lfc_correlation(
    screen,
    guide_idx,
    lfc_label_suffix="",
    method="Pearson",
    cond1="top",
    cond2="bot",
    rep_col="replicate",
    figsize=(7, 7),
    **kwargs,
):
    screen_subset = screen[guide_idx, :]
    lfcs = screen_subset.log_fold_change_reps(
        cond1=cond1, cond2=cond2, rep_col=rep_col, **kwargs
    )
    screen.uns["lfc"] = lfcs
    screen.uns["lfc_corr"] = corr_func_dict[method](lfcs)
    for i, col in enumerate(lfcs.columns):
        rep = col.split(f".{cond1}_{cond2}", 1)[0]
        if "." in rep:
            rep = rep.split(".")
        if isinstance(rep_col, str):
            rep_idx = screen.samples[rep_col].astype(str) == rep
        else:
            rep_idx = (screen.samples[rep_col] == rep).all(axis=1)
        screen.samples.loc[
            rep_idx,
            f"median_lfc_corr.{cond1}_{cond2}",
        ] = np.nanmedian(np.delete(screen.uns["lfc_corr"][i, :], i))
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(
        screen.uns["lfc_corr"],
        ax=ax,
        annot=True,
        xticklabels=lfcs.columns.map(lambda s: s.split(".")[0]),
        yticklabels=lfcs.columns.map(lambda s: s.split(".")[0]),
    )
    ax.set_aspect(1)
    return ax


def G(v):
    bins = np.linspace(0.0, 1.0, 50)
    total = float(np.sum(v))
    yvals = []
    for b in bins:
        bin_vals = v[v <= np.quantile(v, b)]
        bin_fraction = np.sum(bin_vals) / total
        yvals.append(bin_fraction)
    # perfect equality area
    pe_area = np.trapz(bins, x=bins)
    # lorenz area
    lorenz_area = np.trapz(yvals, x=bins)
    gini_val = (pe_area - lorenz_area) / float(pe_area)
    return bins, yvals, gini_val


def plot_guide_coverage(screen, ax=None, figsize=(4, 4)):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    for i, c in enumerate(screen.samples.index):
        sns.kdeplot(
            screen.X[:, i] + 1,
            label=f"{c} (median {np.nanmedian(screen.X[:, i])})",
            clip=(1, screen.X.max() + 10),
            linestyle=linestyles[(i // 10) % 4],
            ax=ax,
        )
    ax.legend(bbox_to_anchor=(1, 1))
    ax.set_xscale("log")
    ax.set_xlabel("counts")
    ax.set_ylabel("# guides")
    ax.set_title("Guide coverage")
    return ax


def plot_X_gini(screen, plot=True):
    gini_vals = []
    if plot:
        plt.figure()
    for c in range(len(screen.samples.index)):
        bins, result, gini_val = G(screen.X[:, c])
        gini_vals.append(gini_val)
        if plot:
            plt.plot(
                bins,
                result,
                label="{} ({:.3f})".format(screen.samples.index[c], gini_val),
            )
    if plot:
        plt.plot(bins, bins, "--", label="perfect eq.")
        plt.xlabel("Cumulative fraction of guides \nfrom lowest to highest read counts")
        plt.ylabel("Cumulative fraction of read counts")
        plt.title("GINI index")
        plt.legend(bbox_to_anchor=(1.02, 1))
    screen.samples["gini_X"] = gini_vals


def get_outlier_guides(
    screen, condit_col: str, mad_z_thres: float = 5, abs_RPM_thres: float = 10000
):
    """Obtain outlier guides.
    For each experimental condition in `samples`, find outlier guides that shows extreme counts compared to guides.
    Outlier guides are defined as those With MAD criteria (z>`mad_z_thres`) and has absolute RPM of `abs_RPM_thres`.
    """

    if "X_RPM" not in screen.layers:
        screen.layers["X_RPM"] = (screen.X / screen.X.sum(axis=0) * 10e6).copy()
    aberr_dict = {}

    for cond in screen.samples[condit_col].unique():
        adata = screen[:, screen.samples[condit_col] == cond]
        median_p = np.nanmedian(adata.layers["X_RPM"].copy(), axis=1)
        aberr_guide_df_condit = []
        for i, sample in enumerate(adata.samples.index):
            outlier_idx = np.where(
                (adata.layers["X_RPM"][:, i] > median_p * mad_z_thres)
                & (adata.layers["X_RPM"][:, i] > abs_RPM_thres)
            )[0]
            outlier_guides = adata.guides.iloc[outlier_idx, :].copy()
            outlier_guides["sample"] = adata.samples.index[i]
            outlier_guides["RPM"] = adata.layers["X_RPM"].copy()[outlier_idx, i]
            aberr_guide_df_condit.append(outlier_guides[["sample", "RPM"]])
        aberr_dict[cond] = aberr_guide_df_condit
    aberr_guide_dfs = []
    for df in aberr_dict.values():
        aberr_guide_dfs.extend(df)
    if len(aberr_guide_dfs) == 0:
        return pd.DataFrame({"name": [], "sample": [], "RPM": []})
    aberr_guides = pd.concat(aberr_guide_dfs, axis=0).reset_index()
    aberr_guides.columns = ["name"] + aberr_guides.columns[1:].tolist()
    # aberr_guides.index = aberr_idx_list
    return aberr_guides



# _read_screen_from_PoolQ.py

__module_name__ = "_read_screen_from_PoolQ.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(
    [
        "vinyard@g.harvard.edu",
    ]
)

# def _assemble_pandas_dict(filepaths, keys=False, sep="\t", **kwargs):

#     """"""

#     PandasDict = {}
#     for i, path in enumerate(filepaths):
#         if keys:
#             df_key = keys[i]
#         else:
#             df_key = keys[i]
#         PandasDict[df_key] = pd.read_csv(path, sep, **kwargs)

#     return PandasDict


# def _find_key_path(outs_dict, key):

#     key_loc = np.where(np.array(list(outs_dict.keys())) == key)[0][0] - 1

#     return outs_dict["paths"][key_loc]


# def _read_poolq_counts_dfs(
#     poolq_outs, counts_keys=["counts", "lognormalized-counts"], **kwargs
# ):

#     """"""

#     filepaths = []
#     for key in counts_keys:
#         filepaths.append(_find_key_path(poolq_outs, key))

#     pd_dict = _assemble_pandas_dict(filepaths, keys=counts_keys, sep="\t", **kwargs)

#     return pd_dict

# def _load_parse_PoolQ_counts_df(
#     PoolQ_OutsDict, ScreenDict=False, counts_keys=["counts", "lognormalized-counts"]
# ):

#     """
#     Load counts df (raw and lognorm) from PoolQ Dict.

#     Parameters:
#     -----------
#     PoolQ_OutsDict

#     ScreenDict

#     counts_keys

#     Returns:
#     --------
#     ScreenDict

#     Notes:
#     ------
#     """

#     if not ScreenDict: # initialize the dict if it does not exist
#         ScreenDict = v.ut.EmptyDict(["layers", "condit_p", "condit_m", "uns"])

#     df_dict = _read_poolq_counts_dfs(PoolQ_OutsDict, counts_keys)

#     ScreenDict["X"] = df_dict["counts"].drop(["Row Barcode", "Row Barcode IDs"], axis=1).values
#     ScreenDict["guides"] = df_dict["counts"][['Row Barcode', 'Row Barcode IDs']]
#     ScreenDict["guides"].columns = ['barcode', 'barcode_id']
#     ScreenDict["layers"]["lognorm_counts"] = df_dict["lognormalized-counts"].drop(["Row Barcode", "Row Barcode IDs"], axis=1)
#     ScreenDict["samples"] = pd.DataFrame(ScreenDict['layers']["lognorm_counts"].columns, columns=["Condition"])

#     return ScreenDict

# def _compose_condit_mp(
#     ScreenDict, PoolQ_OutsDict, counts_keys=["barcode-counts", "unexpected-sequences"]
# ):

#     """
#     Load counts df (raw and lognorm) from PoolQ Dict.

#     Parameters:
#     -----------
#     PoolQ_OutsDict

#     ScreenDict

#     counts_keys

#     Returns:
#     --------
#     ScreenDict

#     Notes:
#     ------
#     """

#     ScreenDict["condit_m"] = _read_poolq_counts_dfs(
#         PoolQ_OutsDict, counts_keys, index_col=[0, 1, -1]
#     )
#     ScreenDict["condit_p"] = _read_poolq_counts_dfs(PoolQ_OutsDict, ["correlation"], index_col=0)

#     return ScreenDict


# def _compose_uns_from_quality_file(ScreenDict, PoolQ_OutsDict, poolq_outs_path, out_path):

#     quality_dfs = _read_poolq_quality_file(
#         poolq_outs_path, analysis_dir=out_path, return_df=True
#     )

#     ScreenDict["uns"]["run_info"] = PoolQ_OutsDict["runinfo"]
#     ScreenDict["uns"]["poolq3"] = PoolQ_OutsDict["poolq3"]

#     for key, _df in quality_dfs.items():
#         ScreenDict["uns"][str(key)] = _df

#     return ScreenDict

# def _read_screen_from_PoolQ(poolq_outs_path, analaysis_out_path="wokrbook.xlsx"):

#     PoolQ_OutsDict = v.ut.glob_dict(poolq_outs_path)

#     ScreenDict = _load_parse_PoolQ_counts_df(PoolQ_OutsDict, ScreenDict=False, counts_keys=["counts", "lognormalized-counts"])
#     ScreenDict = _compose_condit_mp(ScreenDict, PoolQ_OutsDict, counts_keys=["barcode-counts", "unexpected-sequences"])
#     ScreenDict = _compose_uns_from_quality_file(ScreenDict, PoolQ_OutsDict, poolq_outs_path, out_path=analaysis_out_path)

#     return ScreenDict

##################

import pandas as pd
import numpy as np
import os

from ..._utilities._funcs._glob_dict import _glob_dict
from ..._utilities._funcs._FileHandler import _FileHandler


def _read_file_in_dir(dir_path, filename):

    """Useful when you have multiple paths in a dir and you want to reuse the dir name. Uses the FileHandler class."""

    filepath = os.path.join(dir_path, filename)
    f = _FileHandler(filepath=filepath, verbose=False)
    return f.read(return_file=True)


def _create_EmptyDict(keys):

    """Create an empty python Dictionary given a set of keys."""

    Dict = {}
    for key in keys:
        Dict[key] = {}

    return Dict


def _read_poolq_quality_file(
    poolq_outs_path,
    analysis_dir,
    workbook_path="quality.xlsx",
    return_df=False,
    quality_filename="quality.txt",
    poolq_quality_keys=[
        "metadata",
        "bc_readcounts",
        "common_barcodes",
    ],
    verbose=False,
    silent=True,
):

    """"""

    QualityDataFrames = {}

    quality_file = _read_file_in_dir(poolq_outs_path, quality_filename)
    PoolQ_QualityDict = _create_EmptyDict(poolq_quality_keys)

    for n, line in enumerate(quality_file):
        line = _parse_line(line)
        if not line is None:
            if ":" in line[0]:
                PoolQ_QualityDict["metadata"][n] = line[0].split(":")
            elif len(line) > 2:
                _read_count_for_df(PoolQ_QualityDict["bc_readcounts"], n, line)
            elif len(line) == 2:
                _read_count_for_df(PoolQ_QualityDict["common_barcodes"], n, line)
            else:
                if verbose:
                    print("WARNING: LINE NOT ASSIGNED: {}".format(line))

    PoolQ_QualityDict["metadata"]["cols"] = ["metric", "statistic"]
    PoolQ_QualityDict["common_barcodes"]["cols"] = ["barcode", "count"]

    quality_dfs = {}
    for key in poolq_quality_keys:
        quality_dfs[key] = _make_df_with_colnames(PoolQ_QualityDict[key])
    #     v.ut.mkdir_flex(analysis_dir)
    workbook_path = os.path.join(analysis_dir, workbook_path)
    #     v.ut.df_to_excel(
    #         list(quality_dfs.values()),
    #         workbook_path,
    #         sheetnames=poolq_quality_keys,
    #         index=False,
    #         silent=silent,
    #     )

    if return_df:
        return quality_dfs


def _assemble_PoolQ_Quality_Dict():

    return _create_EmptyDict(keys=["metadata", "bc_readcounts", "common_barcodes"])


def _make_df_with_colnames(dictionary, colnames_key="cols"):

    return (
        pd.DataFrame.from_dict(
            dictionary, orient="index", columns=dictionary[colnames_key]
        )
        .drop(colnames_key)
        .reset_index(drop=True)
    )


def _parse_line(line):
    """Parse poolq Quality file lines."""
    if not line == "\n":
        parsed_line = line.strip("\n").split("\t")
        if not parsed_line[0].startswith("Read counts"):
            return parsed_line


def _read_count_for_df(CountDict, n, line):

    """"""

    if len(CountDict) == 0:
        CountDict["cols"] = line
    else:
        CountDict[n] = line

    return CountDict


def _assemble_pandas_dict(filepaths, keys=False, sep="\t", **kwargs):

    """"""

    PandasDict = {}
    for i, path in enumerate(filepaths):
        if keys:
            df_key = keys[i]
        else:
            df_key = keys[i]
        PandasDict[df_key] = pd.read_csv(path, sep, **kwargs)

    return PandasDict


def _find_key_path(outs_dict, key):

    key_loc = np.where(np.array(list(outs_dict.keys())) == key)[0][0] - 1

    return outs_dict["paths"][key_loc]


def _read_poolq_counts_dfs(
    poolq_outs, counts_keys=["counts", "lognormalized-counts"], **kwargs
):

    """"""

    filepaths = []
    for key in counts_keys:
        filepaths.append(_find_key_path(poolq_outs, key))

    pd_dict = _assemble_pandas_dict(filepaths, keys=counts_keys, sep="\t", **kwargs)

    return pd_dict


def _load_parse_PoolQ_counts_df(
    PoolQ_OutsDict, ScreenDict=False, counts_keys=["counts", "lognormalized-counts"]
):

    """
    Load counts df (raw and lognorm) from PoolQ Dict.

    Parameters:
    -----------
    PoolQ_OutsDict

    ScreenDict

    counts_keys

    Returns:
    --------
    ScreenDict

    Notes:
    ------
    """

    if not ScreenDict:  # initialize the dict if it does not exist
        ScreenDict = _create_EmptyDict(["layers", "condit_p", "condit_m", "uns"])

    df_dict = _read_poolq_counts_dfs(PoolQ_OutsDict, counts_keys)

    ScreenDict["X"] = (
        df_dict["counts"].drop(["Row Barcode", "Row Barcode IDs"], axis=1).values
    )
    ScreenDict["guides"] = df_dict["counts"][["Row Barcode", "Row Barcode IDs"]]
    ScreenDict["guides"].columns = ["barcode", "barcode_id"]
    ScreenDict["layers"]["lognorm_counts"] = df_dict["lognormalized-counts"].drop(
        ["Row Barcode", "Row Barcode IDs"], axis=1
    )
    ScreenDict["samples"] = pd.DataFrame(
        ScreenDict["layers"]["lognorm_counts"].columns, columns=["Condition"]
    )

    return ScreenDict


def _compose_condit_mp(
    ScreenDict, PoolQ_OutsDict, counts_keys=["barcode-counts", "unexpected-sequences"]
):

    """
    Load counts df (raw and lognorm) from PoolQ Dict.

    Parameters:
    -----------
    PoolQ_OutsDict

    ScreenDict

    counts_keys

    Returns:
    --------
    ScreenDict

    Notes:
    ------
    """

    ScreenDict["condit_m"] = _read_poolq_counts_dfs(
        PoolQ_OutsDict, counts_keys, index_col=[0, 1, -1]
    )
    ScreenDict["condit_p"] = _read_poolq_counts_dfs(
        PoolQ_OutsDict, ["correlation"], index_col=0
    )

    return ScreenDict


def _compose_uns_from_quality_file(
    ScreenDict, PoolQ_OutsDict, poolq_outs_path, out_path
):

    quality_dfs = _read_poolq_quality_file(
        poolq_outs_path, analysis_dir=out_path, return_df=True
    )

    ScreenDict["uns"]["run_info"] = PoolQ_OutsDict["runinfo"]
    ScreenDict["uns"]["poolq3"] = PoolQ_OutsDict["poolq3"]

    for key, _df in quality_dfs.items():
        ScreenDict["uns"][str(key)] = _df

    return ScreenDict


def _read_screen_from_PoolQ(poolq_outs_path, analaysis_out_path="wokrbook.xlsx"):

    PoolQ_OutsDict = _glob_dict(poolq_outs_path)

    ScreenDict = _load_parse_PoolQ_counts_df(
        PoolQ_OutsDict, ScreenDict=False, counts_keys=["counts", "lognormalized-counts"]
    )
    ScreenDict = _compose_condit_mp(
        ScreenDict,
        PoolQ_OutsDict,
        counts_keys=["barcode-counts", "unexpected-sequences"],
    )
    ScreenDict = _compose_uns_from_quality_file(
        ScreenDict, PoolQ_OutsDict, poolq_outs_path, out_path=analaysis_out_path
    )

    return ScreenDict




from ._annotate_protospacer import _annotate_protospacer

import pandas as pd
import numpy as np

def _add_guide_region_info(df):
    
    df["region_center"] = np.round(
    np.mean(np.stack([df['region_left'].values, df['region_right'].values]), axis=0)
        ).astype(int)
    df["PAM_distance_to_center"] = np.round(abs(df.PAM_loci - df["region_center"]))
    
    return df


def _assign_pam_loci_to_peaks(pam_loci, chromosome_peaks, start_key="Start", end_key="End"):

    """
    Bins PAM loci into peak "bins". DataFrame will still need cleaning after generation. Handled by separate function.
    Parameters:
    -----------
    pam_loci
        numpy array of PAM loci
    chromosome_peaks
        pandas DataFrame containing the start and stop location of each peak.
    Returns:
    --------
    peak_assigned_loci_df
        dataframe matching all PAM loci to a peak
    peak_widths
        loci of peak "bins"
    """
    
    loci_peak_assignment, peak_widths = pd.cut(
        pam_loci,
        np.column_stack((chromosome_peaks[start_key], chromosome_peaks[end_key])).flatten(),
        retbins=True,
    )

    peak_assigned_loci_df = pd.DataFrame(np.stack([pam_loci, loci_peak_assignment]).T)

    return peak_assigned_loci_df, peak_widths

def _reformat_peaks_annotations(pam_loci_df, peak_bins):

    """
    Reformat dataframe of peak-assigned pam loci to a neater df.
    Parameters:
    -----------
    pam_loci_df
        dataframe matching all PAM loci to a peak
    peak_bins
        loci of peak "bins"
    Returns:
    --------
    df
        *cleaned* dataframe matching all PAM loci to a peak
    """

    pam_loci_df = pam_loci_df.dropna()

    left, right = peak_bins[0::2], peak_bins[1::2]

    pams_in_regions_left = np.array([])
    pams_in_regions_right = np.array([])
    index = np.array([])

    for a, i in enumerate(pam_loci_df[1]):

        if i.left in left:
            pams_in_regions_left = np.append(pams_in_regions_left, i.left)
        if i.right in right:
            pams_in_regions_right = np.append(pams_in_regions_right, i.right)
            index = np.append(index, pam_loci_df[0].values[a])

    df = pd.DataFrame(
        np.vstack([index, pams_in_regions_left, pams_in_regions_right]).T,
        columns=["PAM_loci", "region_left", "region_right"],
    )
    
    return df


def _assemble_pam_loci_df(forward, reverse, chromosome_peaks, chromosome_key="Chromosome"):

    """
    Bins PAM loci into peak "bins". Mindful of strand information. Uses adjusted reverse strand loci values.
    Parameters:
    -----------
    forward
        numpy array of forward strand PAM loci
    reverse
        numpy array of reverse strand PAM loci. Should already be (3'-5') adjusted.
    chromosome_peaks
        pandas DataFrame containing the start and stop location of each peak.
    Returns:
    --------
    df
        cleaned DataFrame containing PAM loci, strand assignment, and peak annotation.
    """

 
    forward_pam_peak_assigned, peak_widths_forward = _assign_pam_loci_to_peaks(
        forward, chromosome_peaks
    )

    reverse_pam_peak_assigned, peak_widths_reverse = _assign_pam_loci_to_peaks(
        reverse, chromosome_peaks
    )

    reformated_forward_df = _reformat_peaks_annotations(
        forward_pam_peak_assigned, peak_widths_forward
    )
    reformated_reverse_df = _reformat_peaks_annotations(
        reverse_pam_peak_assigned, peak_widths_reverse
    )

    # add strand info
    reformated_forward_df["strand"] = "+"
    reformated_reverse_df["strand"] = "-"

    df = (
        pd.concat([reformated_forward_df, reformated_reverse_df])
        .sort_values("PAM_loci")
        .reset_index(drop=True)
    )

    df = _add_guide_region_info(df)
    df[chromosome_key] = chromosome_peaks[chromosome_key].unique()[0]
    cols = df.columns.tolist()[-1:] + df.columns.tolist()[:-1]
    df = df[cols]

    df = df.reset_index(drop=True)

    return df

def _make_guide_library_df(forward_pams, reverse_pams, peak_df, chrom_seqs, chromosome_key="Chromosome"):

    """"""

    appended_dfs = []

    for i, chrom in enumerate(peak_df[chromosome_key].unique()):

        df = _assemble_pam_loci_df(
            forward_pams[i],
            reverse_pams[i],
            peak_df.loc[peak_df[chromosome_key].astype(str) == chrom],
        )

        df = _annotate_protospacer(df, chrom_seqs[i])
        appended_dfs.append(df)

    df = pd.concat(appended_dfs)
    df = df.drop_duplicates("protospacer")
    df = df.reset_index(drop=True)

    return df


# _print_screen_object.py
__module_name__ = "_print_screen_object.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(
    [
        "vinyard@g.harvard.edu",
    ]
)


def _print_screen_object(ScreenObject):

    """
    Would be good to return an organizational dictionary
    """

    n_guides, n_conditions = ScreenObject.X.shape[0], ScreenObject.X.shape[1]

    descr = (
        "Genome Editing Screen comprised of n_guides x n_conditions = {} x {}\n".format(
            n_guides, n_conditions
        )
    )

    MainScreenAttributes = [
        "guides",
        "samples",
        "samples_m",
        "samples_p",
        "layers",
        "uns",
    ]

    for attribute in MainScreenAttributes:
        descr += "   {: <11}".format(attribute + ":")
        x = list(ScreenObject.__getattribute__(attribute).keys())
        for n, i in enumerate(x):
            if n != len(x) - 1:
                descr += "'{}', ".format(i)
            else:
                descr += "'{}'".format(i)
        descr += "\n"

    return n_guides, n_conditions, descr




# _get_default_matplotlib_figure_width_height.py
__module_name__ = "_get_default_matplotlib_figure_width_height.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
import matplotlib, numpy


def _get_default_matplotlib_figure_width_height():

    """
    Return default height and width of matplotlib figures.
    
    Parameters:
    -----------
    None
    
    Returns:
    --------
    w, h
        width and height of the default matplotlib figure size
    
    Notes:
    ------
    
    """

    default_wh = matplotlib.rcParams["figure.figsize"]  # w x h
    w, h = default_wh[0], default_wh[1]

    return numpy.array(w, h)


# _utilities __init__.py

__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


#from ._funcs._get_gene_exons import _get_gene_df_from_gtf as get_gene




class _SequenceManipulation:

    """
    Get the complement or reverse compliment of a DNA sequence.

    Parameters:
    -----------
    sequence

    Returns:
    --------
    complement_sequence

    reverse_sequence

    reverse_complement_sequence

    Notes:
    ------
    (1) No dependencies required. Pure python.
    """

    def __init__(self, sequence):

        self.sequence = sequence
        self.ComplimentDict = {"C": "G", "G": "C", "T": "A", "A": "T", "N": "N"}
        self.complement_sequence = ""
        self.reverse_complement_sequence = ""

    def complement(self):

        """
        Get the complement of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        complement_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python.
        """
        for nucleotide in self.sequence:
            self.complement_sequence += self.ComplimentDict[nucleotide]

        return self.complement_sequence

    def reverse(self):

        """
        Get the reverse of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        reverse_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python.
        """

        self.reverse_sequence = self.sequence[::-1]

    def reverse_complement(self):

        """
        Get the reverse complement of a DNA sequence.

        Parameters:
        -----------
        sequence (passed during instantiation).

        Returns:
        --------
        reverse_complement_sequence

        Notes:
        ------
        (1) No dependencies required. Pure python.
        """
        
        ComplimentDict = self.ComplimentDict
        complement_sequence = ""
        for nucleotide in self.sequence:
            complement_sequence += ComplimentDict[nucleotide]

        self.complement_sequence = complement_sequence
        self.reverse_complement_sequence = complement_sequence[::-1]

        return self.reverse_complement_sequence


import scipy.stats
import numpy as np


def _calculate_enrichment_pvalues(arr1, arr2):

    """
    Calculate the probability that two values drawn N times (len(row)) are from distinct distributions.


    Calculate the T-test for the means of two independent samples of scores.
    (source: scipy.stats)


    Notes:
    ------
    (1) This is a two-sided test for the null hypothesis that 2
        independent samples have identical average (expected)
        values.

    (2) This test assumes that the populations have identical
        variances by default.

    (3) Notes (1) and (2) are taken directly from the scipy.stats website.
    """

    p_vals = []

    for row in range(len(arr1)):

        cond1 = arr1[row]
        cond2 = arr2[row]

        p_vals.append(scipy.stats.ttest_ind(cond1, cond2)[1])

    return np.array(p_vals)


def _calculate_baseline_subtraction(condit_1, condit_2, control):

    """Calculate the log fold change relative to the control (i.e., perform baseline subtraction).

    Assumes log-transformed values."""

    condit_1_vs_control = condit_1.values - control.values
    condit_2_vs_control = condit_2.values - control.values

    return condit_1_vs_control, condit_2_vs_control


def _calculate_delta_logfoldchange(condit_1, condit_2):

    """
    Calculate the delta log fold change in guide counts between
    two conditions. Calculates the mean and standard deviation
    over replicates.

    Parameters:
    -----------
    condit_1

    condit_2

    Returns:
    --------


    Notes:
    ------
    (1) Assumes baseline has already been substracted.

    """

    log_fold_change_mean = condit_1.mean(axis=1) - condit_2.mean(axis=1)
    log_fold_change_stdev = (condit_1 - condit_2).std(axis=1)

    return log_fold_change_mean, log_fold_change_stdev



# _readwrite (io) __init__.py

from ._funcs._write_screen_to_csv import _write_screen_to_csv as write_to_csv
from ._funcs._write_screen_to_excel import _write_screen_to_excel as write_to_Excel
from ._funcs._write_experiment_report_to_excel import _write_experiment_report_to_excel as write_experiment_report_to_Excel


# _check_fix_file_extension.py
__module_name__ = "_check_fix_file_extension.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


def _check_fix_file_extension(filepath, extension, silent=False):

    if not filepath.endswith(extension):
        if not silent:
            print(
                "filepath: {} does not have the correct extension: {}\nAppending: {}...".format(
                    filepath, extension, extension
                )
            )
        return filepath + extension
    else:
        return filepath



import pandas as pd

def _add_guide_target_metadata(guide_df, regex_annotations, DirectPairDict=False):

    """"""

    GuideMetadata = {}

    for guide_id in guide_df["barcode_id"].unique():
        if DirectPairDict:
            for key, value in DirectPairDict.items():
                if guide_id == key:
                    GuideMetadata[guide_id] = value
                else:
                    nullcount = 0
                    for annot in regex_annotations:
                        if annot in guide_id:
                            GuideMetadata[guide_id] = annot
                            break
                        else:
                            nullcount += 1
                            if nullcount == 3:
                                GuideMetadata[guide_id] = guide_id

    guide_metadata_df = (
        pd.DataFrame.from_dict(GuideMetadata, orient="index")
        .reset_index()
        .rename({"index": "barcode_id", 0: "target"}, axis=1)
    )

    guide_df = guide_df.merge(guide_metadata_df, on="barcode_id")

    return guide_df



# _glob_dict.py
__module_name__ = "_glob_dict.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])

# package imports #
# --------------- #
import glob, os

# local imports #
# ------------- #
from ._FileHandler import _FileHandler


def _get_basename_no_extension(path):

    try:
        return os.path.basename(path).split(".")[0]
    except:
        return os.path.basename(path)

def _fix_path_for_glob(path):

    if path.endswith("/*"):
        pass
    elif path.endswith("/"):
        path = path + "*"
    else:
        path = path + "/*"

    return path


def _glob_dict(path, verbose=False):

    """
    Loads filepaths from a glob'd directory.

    Parameters:
    -----------
    path
        path to directory containing files.
        type: str

    verbose
        If true, prints FileHandler messages.
        default: False
        type: bool

    Returns:
    --------
    Dict
        keys: filenames
        values: files loaded into memory

    Notes:
    ------
    (1) Can pass with or without "/*"
    """

    filepaths = glob.glob(_fix_path_for_glob(path))

    Dict = {}
    Dict["paths"] = []

    for path in filepaths:
        Dict["paths"].append(path)
        name = _get_basename_no_extension(path)
        file = _FileHandler(filepath=path, verbose=verbose)
        Dict[name] = file.read(return_file=True)

    return Dict


class _GTF(object):
    
    def __init__(self, path):
        
        """"""
        
        self.file = open(path, "r")
        self.header={}
        
        
    def read(path):
        
        """"""
        
        ...


### get context info for sgRNA

import pandas as pd

def _get_reverse_sgRNA_context(seq, df, flank_N):

    """"""

    # isolate all reverse strand sgRNAs
    reverse_guides = df.loc[df.strand == "-"]

    context_bound_5prime = reverse_guides.PAM_loci - [3 + flank_N]
    context_bound_3prime = reverse_guides.PAM_loci + [20 + flank_N]

    guide_context_seqs = []

    if len(context_bound_5prime) == len(context_bound_3prime):
        for i in range(len(context_bound_5prime)):
            guide_context = seq[
                context_bound_5prime.tolist()[i] : context_bound_3prime.tolist()[i]
            ]
            guide_context_seqs.append(guide_context)

    reverse_guides["guide_context"] = guide_context_seqs

    return reverse_guides


def _get_forward_sgRNA_context(seq, df, flank_N):

    """"""

    # isolate all reverse strand sgRNAs
    forward_guides = df.loc[df.strand == "+"]

    context_bound_5prime = forward_guides.PAM_loci - [20 + flank_N]
    context_bound_3prime = forward_guides.PAM_loci + [3 + flank_N]

    guide_context_seqs = []

    if len(context_bound_5prime) == len(context_bound_3prime):
        for i in range(len(context_bound_5prime)):
            guide_context = seq[
                context_bound_5prime.tolist()[i] : context_bound_3prime.tolist()[i]
            ]
            guide_context_seqs.append(guide_context)

    forward_guides["guide_context"] = guide_context_seqs

    return forward_guides


def _add_sgRNA_context(df, chrom_seq, flank_N=10):

    """
    Get the flanking N nts (default N=10) around a protospacer + PAM
    DF and chrom_seqs must be sorted by chromosome order
    """

    dfs = []

    for i, chrom in enumerate(df["Chromosome"].unique()):

        chrom_df = df.loc[df["Chromosome"] == chrom]

        forw_context_added_df = _get_forward_sgRNA_context(
            chrom_seq[i], chrom_df, flank_N
        )
        rev_context_added_df = _get_reverse_sgRNA_context(
            chrom_seq[i], chrom_df, flank_N
        )

        dfs.append(forw_context_added_df)
        dfs.append(rev_context_added_df)

    df = pd.concat(dfs)

    return df



# _get_chromosome_sequence.py
__module_name__ = "_get_chromosome_sequence.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])

# package imports #
# --------------- #


def _get_chromosome_sequence(ref_seq_path, query_chr):

    """
    Get a specific chromosome sequence from a reference genome. Also report the length of that sequence.
    Parameters:
    -----------
    ref_seq_path
        Path to a reference genome fasta file.
    query_chr
        Chromosome sequence to isolate
        type: str

    Returns:
    --------
    chromosome_reference_seq
        type: str

    Notes:
    ------
    """

    for record in SeqIO.parse(ref_seq_path, "fasta"):

        if record.description.split()[0] == query_chr:
            chromosome_reference_seq = str(record.seq)

            return chromosome_reference_seq




from ._GTF_Module import _GTF
from ._get_screen_target_features import _gene_from_gtf, _get_feature, _get_gene_body_bounds
from ._identify_overlapping_fragments import _OverlappingFragments
from ._query_region_for_PAM import _query_region_for_PAM

from ._make_guide_library_df import _make_guide_library_df
from ._annotate_protospacer import _annotate_protospacer
from ._add_sgRNA_context import _add_sgRNA_context
from ._scan_for_BsmbI import _scan_for_BsmbI

from ._widen_regions import _widen_regions
from ._add_region_labels import _add_region_labels

class _ScreeningTarget:
    
    def __init__(self, GTF, ref_seq_path):
        
        self.ref_seq_path = ref_seq_path
        
        if type(GTF) == str:
            gtf = _GTF(GTF)
            self.gtf_file = gtf.file
        else:
            self.gtf_file = GTF.file
        
    def get_gene(self, chromosome, gene):
        
        self.gene = gene
        self.gene_df = _gene_from_gtf(self.gtf_file, chromosome, gene)
        self.gene_min, self.gene_max = _get_gene_body_bounds(self.gene_df)
        
    def get_features(self, feature):
        
        self.feature_df = _get_feature(self.gene_df, feature)

    def get_exons(self):
        
        
        self.exon_df = _get_feature(self.gene_df, feature="exon")
        self.exon_df = self.exon_df[[0, 3, 4]]
        self.exon_df.columns = ["Chromosome", "Start", "End"]
        
#         print("\nCheck the exon sequences for overlaps...\n")
        _exon_overlaps = _OverlappingFragments(self.exon_df)
        _exon_overlaps.find_all()
        
        self.exon_df = _exon_overlaps.df
        self.exon_df['Start'] = self.exon_df['Start'].astype(int)
        self.exon_df['End'] = self.exon_df['End'].astype(int)
        self.exon_df['exon_length'] = self.exon_df["End"].astype(int) - self.exon_df["Start"].astype(int)
        
        
    def build_guide_library(self, exons=True, feature=False, PAM="NGG", filter_BsmbI=True, widen=False):
        
        """
        forward_pams, reverse_pams, KMT2C, chrom_seqs
        """
        
        if exons == True:
            print("\nUsing {} exons...\n".format(self.gene))
            self.regions_df = self.exon_df
        else:
            try:
                self.regions_df = self.feature_df
            except:
                self.regions_df = self._get_feature(self.gene_df, feature)
        if widen:
            print("\nWidening regions by {} nucleotides on either side of each boundary...\n".format(widen))
            self.regions_df = _widen_regions(self.regions_df, widen)
                
        print("\nBuilding sgRNA library DataFrame...")
        
        forward_pams, reverse_pams, chrom_seq = _query_region_for_PAM(self.regions_df, self.ref_seq_path, PAM=PAM)
        
        print("PAMs, Chromosome Seqeunce isolated")
        
        self.guide_df = _make_guide_library_df(forward_pams, reverse_pams, self.regions_df, chrom_seq)
        
        print("\nAdding sgRNA sequence context")
        self.guide_df = _add_sgRNA_context(self.guide_df, chrom_seq)
#         self.guide_df = _add_region_labels(self.guide_df)
                
        print("\nAdding BsmbI annotations")
        self.guide_df, self.BsmbI_guide_df = _scan_for_BsmbI(self.guide_df, filter_BsmbI)
        self.guide_df = self.guide_df.reset_index(drop=True)
        
    def write(filename=False):
        
        if not filename:
            filename = "perturb_tools.target_library.{}.xlsx".format(self.gene)

        self.guide_df.to_excel(filename)


def _read_poolq_quality_file(
    poolq_outs_path,
    analysis_dir,
    workbook_path="quality.xlsx",
    return_df=False,
    quality_filename="quality.txt",
    poolq_quality_keys=[
        "metadata",
        "SampleBarcodeReadCounts",
        "CommonSampleBarcodeReadCounts",
    ],
    verbose=False,
    silent=True,
):

    """"""

    QualityDataFrames = {}

    quality_file = _read_file_in_dir(poolq_outs_path, quality_filename)
    PoolQ_QualityDict = _create_EmptyDict(poolq_quality_keys)

    for n, line in enumerate(quality_file):
        line = _parse_line(line)
        if not line is None:
            if ":" in line[0]:
                PoolQ_QualityDict["metadata"][n] = line[0].split(":")
            elif len(line) > 2:
                _read_count_for_df(
                    PoolQ_QualityDict["SampleBarcodeReadCounts"], n, line
                )
            elif len(line) == 2:
                _read_count_for_df(
                    PoolQ_QualityDict["CommonSampleBarcodeReadCounts"], n, line
                )
            else:
                if verbose:
                    print("WARNING: LINE NOT ASSIGNED: {}".format(line))

    PoolQ_QualityDict["metadata"]["cols"] = ["metric", "statistic"]
    PoolQ_QualityDict["CommonSampleBarcodeReadCounts"]["cols"] = ["barcode", "count"]

    quality_dfs = {}
    for key in poolq_quality_keys:
        quality_dfs[key] = _make_df_with_colnames(PoolQ_QualityDict[key])
    v.ut.mkdir_flex(analysis_dir)
    workbook_path = os.path.join(analysis_dir, workbook_path)
    _write_df_to_excel(
        list(quality_dfs.values()),
        workbook_path,
        sheetnames=poolq_quality_keys,
        index=False,
        silent=silent,
    )

    if return_df:
        return quality_dfs
    
    
def _assemble_PoolQ_Quality_Dict():

    return _create_EmptyDict(
        keys=["metadata", "bc_counts", "CommonSampleBarcodeReadCounts"]
    )


def _make_df_with_colnames(dictionary, colnames_key="cols"):

    return (
        pd.DataFrame.from_dict(
            dictionary, orient="index", columns=dictionary[colnames_key]
        )
        .drop(colnames_key)
        .reset_index(drop=True)
    )


def _read_file_in_dir(dir_path, filename):

    """Useful when you have multiple paths in a dir and you want to reuse the dir name. Uses the FileHandler class."""

    filepath = os.path.join(poolq_outs_path, quality_filename)
    f = FileHandler(filepath=filepath, verbose=False)
    return f.read(return_file=True)

def _read_file_in_dir(dir_path, filename):

    """Useful when you have multiple paths in a dir and you want to reuse the dir name. Uses the FileHandler class."""

    filepath = os.path.join(poolq_outs_path, quality_filename)
    f = FileHandler(filepath=filepath, verbose=False)
    return f.read(return_file=True)


def _parse_line(line):
    """Parse poolq Quality file lines."""
    if not line == "\n":
        parsed_line = line.strip("\n").split("\t")
        if not parsed_line[0].startswith("Read counts"):
            return parsed_line


def _read_count_for_df(CountDict, n, line):

    """"""

    if len(CountDict) == 0:
        CountDict["cols"] = line
    else:
        CountDict[n] = line

    return CountDict


# _get_chromosome_sequence.py

from Bio import SeqIO


def _get_chromosome_sequence(ref_seq_path, query_chr, return_length=False):

    """
    Get a specific chromosome sequence from a reference genome. Also report the length of that sequence.
    
    Parameters:
    -----------
    ref_seq_path [ required ]
        Path to a reference genome fasta file.
    
    query_chr [ required ]
        Chromosome sequence to isolate
        type: str
        
    return_length [ optional ]
        default: False
        type: bool
        
    Returns:
    --------
    chromosome_reference_seq
        type: str
    
    len(chromosome_reference_seq) [ optional ]
    
    Notes:
    ------
    (1) if return_length is true, a list is returned to avoid setting two outputs.  
    """

    for record in SeqIO.parse(ref_seq_path, "fasta"):

        if record.description.split()[0] == query_chr:
            chromosome_reference_seq = str(record.seq)

        if return_length:
            return [chromosome_reference_seq, len(chromosome_reference_seq)]
        else:
            return chromosome_reference_seq




import setuptools
from setuptools import setup
import re
import os
import sys


setup(
    name="perturb-tools",
    version="0.3.5",
    python_requires=">3.7.0",
    author=[
        "Michael E. Vinyard - Harvard University - Massachussetts General Hospital - Broad Institute of MIT and Harvard",
        "Jayoung K. Ryu - Harvard University - Massachussetts General Hospital - Broad Institute of MIT and Harvard",
    ],
    author_email="mvinyard@broadinstitute.org, jayoung_ryu@g.harvard.edu",
    url="https://github.com/pinellolab/perturb-tools",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="perturb-tools - Analysis Framework for Pooled CRISPR Genome Editing Screens.",
    packages=setuptools.find_packages(),
    #    packages=[
    #        "perturb_tools",
    #        "perturb_tools._arithmetic",
    #        "perturb_tools._experimental_design",
    #        "perturb_tools._external_tools",
    #        "perturb_tools._framework",
    #        "perturb_tools._framework._supporting_functions",
    #        "perturb_tools._normalization",
    # 	"perturb_tools._plotting",
    # 	"perturb_tools._readwrite",
    # 	"perturb_tools._utilities",
    #    ],
    install_requires=[
        "matplotlib>=3.4",
        "anndata>=0.7.1",
        "numpy>=1.19.2",
        "pandas>=1.1.2",
        "biopython>=1.79",
        "plotly",
        "regex",
        "xlsxwriter",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT",
)




# _FileHandler.py
__module_name__ = "_FileHandler.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])

class _FileHandler:
    
    def __init__(self, filepath=False, verbose=True):
        
        """Uses only pure python"""
        
        self.verbose = verbose
        
        if filepath:
            self.filepath = filepath
    
    def read(self, return_file=False, filepath=False):
        
        """
        """
        
        if filepath:
            self.filepath = filepath
        if self.verbose:
            print("Reading: {}".format(self.filepath))
        with open(self.filepath, "r") as f:
            self.file = f.readlines()
        
        if return_file:
            return self.file
        
    def count(self):
        
        self.file_length = len(self.file)
        if self.verbose:
            print(self.file_length)



import pandas as pd

from ._supporting_functions._annotate_protospacer import _annotate_protospacer
from ._supporting_functions._add_guide_target_metadata import _add_guide_target_metadata
from ._supporting_functions._add_context_sequence_to_gene_dict import _add_context_sequence_to_gene_dict

class _GuideAnnotation:
    def __init__(self, guide_df, genes, chromosome, start, stop):

        """"""

        self.guide_df = _annotate_protospacer(guide_df)
        self.chr = chromosome
        self.start = start
        self.stop = stop
        #self.GeneDict = _create_gene_dict(genes, chromosome, start, stop)
        self.gene_df = pd.DataFrame.from_dict(self.GeneDict).T

    def add_target_metadata(self, regex_annotations=None, DirectPairDict=False):

        """Added annotations should match the gene target used for downstream analysis."""

        if regex_annotations == None:
            regex_annotations = self.genes
        elif regex_annotations:
            regex_annotations = regex_annotations

        self.guide_df = _add_guide_target_metadata(
            self.guide_df, regex_annotations, DirectPairDict
        )

    def fetch_sequence_context(self, ref_seq_path):

        """"""

        self.GeneDict = _add_context_sequence_to_gene_dict(
            self.GeneDict, self.gene_df, ref_seq_path
        )

    def annotate_position(self):

        """ """
        pass
        #self.guide_df = _annotate_guide_position(self.GeneDict, self.guide_df)
        
        
        
def _annotate_sgRNAs(guides, genes, chrom, start, stop, annotations, DirectPairDict, ref_seq_path):
    
    """Given the metadata csv with targets and Mb positions, annotate the sgRNAs in the screen.genes table."""
    
    
    _guides = _GuideAnnotation(guides, genes, chrom, start, stop)
    _guides.add_target_metadata(annotations, DirectPairDict)
    _guides.fetch_sequence_context(ref_seq_path)
    _guides.annotate_position()
    
    return _guides.guide_df



# _plot_correlation_heatmap.py
__module_name__ = "_plot_correlation_heatmap.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas as pd
import os


# local imports #
# ------------- #
from ._get_default_matplotlib_figure_width_height import _get_default_matplotlib_figure_width_height
from ._set_matplotlib_rc_params import _set_matplotlib_rc_params


# setup figure params
_set_matplotlib_rc_params()


def _plot_correlation_heatmap(
    df, title="Sample Correlation", figsize=1, title_y=1.15, title_x=0, savename=False, cbar_aspect=20, cbar_shrink=0.6,
):

    default_h, default_w = _get_default_matplotlib_figure_width_height()
    
    fig = plt.figure(figsize=(default_w * figsize, default_h * figsize))
    gridspec = GridSpec(1, 1)
    ax = fig.add_subplot(gridspec[0, 0])
    im = ax.imshow(df.values.astype(float), cmap="plasma", vmin=0, vmax=1)
    plt.xticks(
        range(len(df.columns.values.astype(str))),
        df.columns.values.astype(str),
        rotation=60,
        ha="left",
    )
    ax.xaxis.tick_top()
    plt.yticks(
        range(len(df.columns.values.astype(str))),
        df.columns.values.astype(str),
        rotation=0,
    )
    plt.colorbar(im, shrink=cbar_shrink, aspect=cbar_aspect)
    plt.title(title, y=title_y, x=title_x, fontsize=16)

    if savename:
        figsavename = savename + ".png"
        plt.savefig(figsavename, bbox_inches="tight")
    plt.tight_layout()
    plt.show()
    if savename:
        return figsavename






import pandas as pd

from ..._experimental_design._funcs._GTF_Module import _GTF

def _get_gene_df_from_gtf(gtf_file, gene, chromosome):

    """"""

    GeneDict = {}

    n = 0
    for i, line in enumerate(gtf_file.readlines()):
        line = line.strip("\n").split("\t")
        if line[0] == chromosome:
            line_ = line[8].split(" ")
            if line_[10] == "gene_name":
                if line_[11].startswith('"{}'.format(gene)):
                    GeneDict[n] = line
                    n += 1
    df = pd.DataFrame.from_dict(GeneDict, orient="index")

    return df


def _get_feature(df, feature="exon"):

    """"""

    df_ = (
        df.loc[df[2] == feature]
        .loc[df[1] == "HAVANA"]
        .reset_index(drop=True)[[0, 3, 4]]
    )
    df_.columns = ["Chromosome", "Start", "End"]
    return df_


def _get_gene_exons(gene, chromosome, gtf_path):

    """"""

    gtf = _GTF(gtf_path)
    gene_df = _get_feature(_get_gene_df_from_gtf(gtf.file, gene, chromosome))

    return gene_df


# plotting __init__.py

# from ._funcs._plot_guide_enrichment import _plot_guide_enrichment
from ._funcs._get_default_matplotlib_figure_width_height import *

# from ._funcs._plot_rep_consistency import *
from ._funcs._plot_correlation_heatmap import *
from ._funcs._set_matplotlib_rc_params import *
from ._funcs._plot_stats import *



# _write_screen.py
__module_name__ = "_write_screen.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(
    [
        "vinyard@g.harvard.edu",
    ]
)


# package imports #
# --------------- #
import os


# local imports #
# ------------- #
from ..._utilities._funcs._flexible_mkdir import _flexible_multilevel_mkdir


def _save_dict_of_dfs(df_dict, out_path, filename, verbose=False):

    for key in df_dict.keys():
        file_outpath = os.path.join(out_path, "{}.{}.csv".format(filename, key))
        if verbose:
            print(file_outpath)
        df_dict[key].to_csv(file_outpath)


def _write_screen_to_csv(screen, out_path="CRISPR_screen"):

    """This function will eventually be replaced with something more native / similar to AnnData."""

    _flexible_multilevel_mkdir(out_path)

    screen.X.tofile(
        os.path.join(out_path, "screen.X.csv"),
        sep=",",
    )
    screen.guides.to_csv(os.path.join(out_path, "screen.guides.csv"))
    screen.samples.to_csv(os.path.join(out_path, "screen.samples.csv"))
    _save_dict_of_dfs(screen.samples_m, out_path, "screen.samples_m")
    _save_dict_of_dfs(screen.samples_p, out_path, "screen.samples_p")
    _save_dict_of_dfs(screen.layers, out_path, "screen.layers")



# main __init__.py


from . import _arithmetic as calc
from . import _experimental_design as design
from . import _external_tools as tools
from . import _normalization as norm
from . import _plotting as pl
from . import _qc as qc
from . import _readwrite as io
from . import _utilities as util
from ._framework._ScreenModule import _Screen as Screen
from ._framework._ScreenModule import concat, read_csv, read_h5ad



# _write_multi_df_to_excel.py
__module_name__ = "_write_multi_df_to_excel.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(
    [
        "vinyard@g.harvard.edu",
    ]
)

# package imports #
# --------------- #
import pandas as pd

# local imports #
# ------------- #
from ._check_fix_file_extension import _check_fix_file_extension


def _append_dfs_and_name_from_dict(df_dict, df_list, sheet_names):

    subclass = "df_dict={}".format(df_dict).split("=")[0]

    for key in df_dict.keys():
        df_list.append(df_dict[key])
        sheet_names.append(".".join([subclass, key]))
    return df_list, sheet_names


def _collect_screen_dfs(screen, include_uns=False):

    """collect all main elements of a screen as a pandas DataFrame."""

    df_list = [
        pd.DataFrame(screen.X, index=screen.guides.index, columns=screen.samples.index)
    ]
    sheet_names = ["X"]
    for key, mat in screen.layers.items():
        df_list.append(
            pd.DataFrame(mat, index=screen.guides.index, columns=screen.samples.index)
        )
        sheet_names.append(key)

    df_list.extend([screen.guides, screen.samples])
    sheet_names.extend(["guides", "samples"])

    for i in [screen.samples_m, screen.samples_p]:
        df_list, sheet_names = _append_dfs_and_name_from_dict(i, df_list, sheet_names)

    if include_uns:
        for key in screen.uns.keys():
            df_list.append(pd.DataFrame(screen.uns[key]))
            sheet_names.append(".".join(["screen.uns", key]))

    return df_list, sheet_names


def _write_screen_to_excel(
    screen,
    workbook_path="./CRISPR_screen.xlsx",
    index=True,
    silent=False,
    include_uns=False,
):

    """
    Write one or more pandas.DataFrames to individual sheets of an Excel Workbook (workbook.xlsx).

    Parameters:
    -----------
    df_list
        type: list(pd.DataFrame, ...)

    workbook_path
        type: str

    index
        default: False
        type: bool

    silent
        default: False
        type: bool

    Returns:
    --------
    None
        writes to excel workbook (.xlsx)

    Notes:
    ------
    (1)
    """

    df_list, sheetnames = _collect_screen_dfs(screen, include_uns=include_uns)
    workbook_path = _check_fix_file_extension(workbook_path, ".xlsx", silent)

    assert len(df_list) == len(sheetnames), print(
        "Length of df_list must equal the length of sheet names passed."
    )

    with pd.ExcelWriter(workbook_path) as writer:

        if not silent:
            print("Writing to: {}\n".format(workbook_path))

        for n, df_ in enumerate(df_list):
            if not sheetnames:
                sheet = "Sheet_{}".format(n)
            else:
                sheet = sheetnames[n]
            if not silent:
                print("\tSheet {}:\t{}".format(str(int(n + 1)), sheet))
            try:
                df_.to_excel(writer, sheet, index=index)
            except ValueError as e:
                print(
                    "While writing {}, following error occurred: passing the data.".format(
                        n
                    )
                )
                print(e)
                continue




import numpy as np

def _get_gene_y_position(lfc):
    return lfc.min() - (lfc.max() - lfc.min()) / 10

def _plot_gene_single_exon(fig, start, end, y, fillcolor="red", opacity=0.8):

    """"""

    fig.add_shape(
        x0=start,
        x1=end,
        y0=y-0.3,
        y1=y,
        line_width=0,
        fillcolor=fillcolor,
        opacity=opacity,
    )

def _plot_gene_exon(fig, gene, gene_df, lfc, exon_color):

    """"""
    
    exon_plot_y = _get_gene_y_position(lfc)
    
    _start, _end = gene_df.Start.min(), gene_df.End.max()
    exon_plot_x = np.mean([float(_start), float(_end)])
    
    fig.add_annotation(x=exon_plot_x, y=exon_plot_y + 0.15, text=gene, showarrow=False,)
    fig.update_layout(showlegend=False)
    
    fig.add_shape(
        x0=_start,
        x1=_end,
        y0=exon_plot_y-0.18,
        y1=exon_plot_y-.12,
        line_width=0,
        fillcolor="red",
        opacity=1,
    )
    
    
    for exon in range(len(gene_df)):

        _plot_gene_single_exon(
            fig,
            gene_df.Start[exon],
            gene_df.End[exon],
            y=exon_plot_y,
            fillcolor=exon_color,
            opacity=0.25,
        )



import pandas as pd
#import vintools as v
import re, numpy as np

def _transform_guide_spanning_dict_to_guide_df_addition(GuidesSpanning, guide_df):

    GuidePosition = {}
    for strand in ["+", "-"]:

        for protospacer, value in GuidesSpanning[strand].items():

            GuidePosition[protospacer] = [strand, value[0], value[1], value[2]]

    df_ = (
        pd.DataFrame(GuidePosition, index=["strand", "Chromosome", "Start", "End"])
        .T.reset_index()
        .rename({"index": "protospacer"}, axis=1)
    )
    
    guide_df = guide_df.merge(df_, on="protospacer", how="outer")
    guide_df["center"] = np.mean([guide_df["Start"], guide_df["End"]], axis=0)
    
    return guide_df

def _annotate_guide_position(GeneDict, guide_df):
        
    GuidesSpanning = v.ut.EmptyDict(["+", "-"])
    GuidesSpanning["+"] = {}
    GuidesSpanning["-"] = {}
    non_matching = []

    for target in GeneDict.keys():
        
        pos_target = GeneDict[target]["seq"]["+"]
        neg_target = GeneDict[target]["seq"]["-"]
        
        true_start = GeneDict[target]["Start"]*1e6
        true_end = GeneDict[target]["End"]*1e6

        target_guides = guide_df.loc[guide_df["target"] == target]
        
        for spacer in target_guides.protospacer.values:
            try:
                span = re.search(spacer, pos_target).span()
                guide_start, guide_end = span[0] + true_start, span[1] + true_start
                GuidesSpanning["+"][spacer] = (GeneDict[target]["Chromosome"], int(guide_start), int(guide_end))
            except:
                try:
                    span = re.search(spacer, neg_target).span()
                    guide_start, guide_end = true_end - span[0], true_end - span[1]
                    GuidesSpanning["-"][spacer] = (GeneDict[target]["Chromosome"], int(guide_start), int(guide_end))
                except:
                    non_matching.append([spacer, target])

    if len(non_matching) > 0:
        print("Number of non-mapping guides: {}".format(len(non_matching)))

    return _transform_guide_spanning_dict_to_guide_df_addition(GuidesSpanning, guide_df)